import os

from naeural_core import constants as ct

class ConfigCommandHandlers:
  
  def update_config_stream(self, delta_config_stream, initiator_id, session_id, verbose=1):
    stream_name = delta_config_stream[ct.CONFIG_STREAM.NAME]
    validated_command = delta_config_stream.get(ct.COMMS.COMM_RECV_MESSAGE.K_VALIDATED, False)
    # we pop the sender address and we will use it for initiator and/or modified-by
    sender_addr = delta_config_stream.pop(ct.COMMS.COMM_RECV_MESSAGE.K_SENDER_ADDR, None)
    # now we pop the initiator Id and decide if we should use it or use it as modified-by id
    sender_id = delta_config_stream.pop(ct.CONFIG_STREAM.INITIATOR_ID, None)
    # now just a sanity check
    if sender_id != initiator_id:
      msg = "ERROR: Received initiator inconsistent with the command for pipeline '{}'. Initiator: {} vs Sender: {}".format(
        stream_name, initiator_id, sender_id
      )
      self.P(msg, color='r')
      self._create_notification(
        notif=ct.STATUS_TYPE.STATUS_ABNORMAL_FUNCTIONING, 
        notif_code=ct.NOTIFICATION_CODES.PIPELINE_FAILED,
        initiator_id=initiator_id,
        session_id=session_id,
        stream_name=stream_name,
        msg=msg,
      )
      return
    
    if stream_name.lower() == self.admin_pipeline_name.lower():
      msg = "WARNING: Attempted detected to modify whole administration pipeline. You can only use update_pipeline_instance to modify individual plugins."
      self.P(msg, color='r')
      self._create_notification(
        notif=ct.STATUS_TYPE.STATUS_ABNORMAL_FUNCTIONING, 
        notif_code=ct.NOTIFICATION_CODES.PIPELINE_DCT_CONFIG_FAILED,
        initiator_id=initiator_id,
        session_id=session_id,
        stream_name=stream_name,
        msg=msg,
      )
      return     
    #endif admin pipeline modification is not allowed 
      

    is_duplicate = self._check_duplicate_last(
      payload=delta_config_stream,
      payload_type=ct.PAYLOAD_CT.COMMANDS.UPDATE_CONFIG,
      
    )
    if is_duplicate:
      msg = "Received duplicate {} for pipeline {} from {}. In future this will be skipped.".format(
        ct.PAYLOAD_CT.COMMANDS.UPDATE_CONFIG,
        stream_name, initiator_id,
      )
      self.P(msg, color='error')
      self._create_notification(
        notif=ct.STATUS_TYPE.STATUS_ABNORMAL_FUNCTIONING, 
        notif_code=ct.NOTIFICATION_CODES.PIPELINE_FAILED,
        msg=msg, 
        stream_name=stream_name,
        session_id=session_id,
        initiator_id=initiator_id,
        displayed=True,
      )
      # return
    #endif

    config_stream = self.dct_config_streams.get(stream_name, {})
    is_new = False
    if len(config_stream) > 0:
      msg = "Received {} command to update for existing pipeline '{}' ({}) from {}:{}".format(
        "VALIDATED" if validated_command else "NON-VALIDATED", 
        stream_name, config_stream.get(ct.CONFIG_STREAM.TYPE),
        initiator_id,
        session_id
      )
    else:
      msg = "Received {} command for new pipeline '{}' from {}:{}".format(
        "VALIDATED" if validated_command else "NON-VALIDATED", 
        stream_name,
        initiator_id,
        session_id
      )
      is_new = True
    #endif
    self.P(msg)
    self._create_notification(
      notif=ct.STATUS_TYPE.STATUS_NORMAL, 
      msg=msg, 
      initiator_id=initiator_id,
      session_id=session_id,
      stream_name=None,
      displayed=True,
    )
    
    if is_new:
      delta_config_stream[ct.CONFIG_STREAM.K_INITIATOR_ADDR] = sender_addr
      delta_config_stream[ct.CONFIG_STREAM.K_INITIATOR_ID] = initiator_id
    else:
      previous_initiator_id = config_stream.get(ct.CONFIG_STREAM.INITIATOR_ID)
      previous_initiator_addr = config_stream.get(ct.CONFIG_STREAM.K_INITIATOR_ADDR)
      if previous_initiator_id != initiator_id or previous_initiator_addr != sender_addr:
        msg = "Received update for '{}' from a different initiator <{}:{}> than original <{}:{}>".format(
          stream_name,           
          initiator_id, sender_addr,
          previous_initiator_id, previous_initiator_addr,          
        )
    #endif
    # now set the modified-by fields
    delta_config_stream[ct.CONFIG_STREAM.K_MODIFIED_BY_ADDR] = sender_addr
    delta_config_stream[ct.CONFIG_STREAM.K_MODIFIED_BY_ID] = initiator_id
    
    config_stream = self._apply_delta_to_config(config_stream, delta_config_stream)
    config_stream = self.keep_good_stream(config_stream)

    if config_stream is not None:
      stream_name = config_stream[ct.CONFIG_STREAM.NAME]
      self._save_stream_config(config_stream)
      self.dct_config_streams[stream_name] = config_stream
      
      str_new = "new" if is_new else "existing"
      msg = "Successfully updated {} configuration for pipeline '{}' initiated by '{}'".format(str_new, stream_name, initiator_id)
      if verbose >= 1:
        self.P(msg, color='g')
        
      if session_id is None:
        session_id = self.dct_config_streams[stream_name].get(ct.CONFIG_STREAM.SESSION_ID)
      if initiator_id is None:
        initiator_id = self.dct_config_streams[stream_name].get(ct.CONFIG_STREAM.INITIATOR_ID)
        
      self._create_notification(
        notif=ct.STATUS_TYPE.STATUS_NORMAL, 
        notif_code=ct.NOTIFICATION_CODES.PIPELINE_OK, # if not is_new else None, # REMOVED: always ok to send notif on success
        msg=msg,
        stream_name=stream_name, # this only confirms the config NOT the actual running
        session_id=session_id,
        initiator_id=initiator_id,
        # displayed=verbose >= 1,
      )      
    else:
      self._create_notification(
        notif=ct.STATUS_TYPE.STATUS_ABNORMAL_FUNCTIONING, 
        notif_code=ct.NOTIFICATION_CODES.PIPELINE_FAILED,
        msg=msg,
        stream_name=stream_name, # this only confirms the config NOT the actual running
        session_id=session_id,
        initiator_id=initiator_id,
        # displayed=verbose >= 1,
      )      
    #endif
    return    
  

  def update_pipeline_instance(self, dct_config_data, initiator_id, session_id):
    pipeline_name = dct_config_data.get(ct.PAYLOAD_DATA.NAME)
    signature = dct_config_data.get(ct.PLUGIN_INFO.SIGNATURE)
    instance_id = dct_config_data.get(ct.PLUGIN_INFO.INSTANCE_ID)
    instance_update = dct_config_data.get(ct.PAYLOAD_DATA.INSTANCE_CONFIG)
    validated_command = dct_config_data.get(ct.COMMS.COMM_RECV_MESSAGE.K_VALIDATED, False)
    sender_addr = dct_config_data.get(ct.COMMS.COMM_RECV_MESSAGE.K_SENDER_ADDR, None)
    if not (isinstance(instance_update, dict) and len(instance_update) > 0):
      instance_update = None
    incomplete_request = pipeline_name is None or signature is None or instance_id is None or instance_update is None
    non_existant_stream = pipeline_name is not None and pipeline_name not in self.dct_config_streams
    notif_code = None
    notif = None
    if incomplete_request or non_existant_stream:
      msg = "Received invalid '{}' command for {}:{}:{}. Incomplete request: {}, Non existant stream: {}".format(
        ct.PAYLOAD_CT.COMMANDS.UPDATE_PIPELINE_INSTANCE,
        pipeline_name, signature, instance_id,
        incomplete_request, non_existant_stream,
      )
      notif = ct.STATUS_TYPE.STATUS_EXCEPTION
      notif_code=ct.NOTIFICATION_CODES.PIPELINE_FAILED
      info = "Command should contain NAME, SIGNATURE, INSTANCE_ID as well as INSTANCE_CONFIG 'delta' object and MUST address a existing pipeline. Full received command: {}".format(dct_config_data)
    else:
      
      if session_id is None:
        session_id = self.dct_config_streams[pipeline_name].get(ct.PAYLOAD_DATA.SESSION_ID)
      else:
        self.dct_config_streams[pipeline_name][ct.PAYLOAD_DATA.SESSION_ID] = session_id

      if initiator_id is None:
        initiator_id = self.dct_config_streams[pipeline_name].get(ct.PAYLOAD_DATA.INITIATOR_ID)    
      else:
        # update modified-by fields
        self.dct_config_streams[pipeline_name][ct.CONFIG_STREAM.K_MODIFIED_BY_ID] = initiator_id
      #endif
      self.dct_config_streams[pipeline_name][ct.CONFIG_STREAM.K_MODIFIED_BY_ADDR] = sender_addr
        
        
      instance_config = self._get_plugin_instance(
        stream_name=pipeline_name, signature=signature, instance_id=instance_id
      )
      if instance_config is not None:
        # the actual update
        instance_config = self._apply_delta_to_config(instance_config, instance_update, ignore_fields=[ct.PLUGIN_INFO.INSTANCE_ID])
        # end update
        self._save_stream_config(self.dct_config_streams[pipeline_name])
        notif = ct.STATUS_TYPE.STATUS_NORMAL
        notif_code=ct.NOTIFICATION_CODES.PIPELINE_OK
        msg = "Successfully updated with {} command {}:{}:{}".format(
          "VALIDATED" if validated_command else "NON-VALIDATED (sign/hash issue)", instance_id, pipeline_name, signature
        )
        info = ""
      else:
        msg = "Cannot find instance '{}' in pipeline/plugin <{}:{}>".format(instance_id, pipeline_name, signature)
        info = "Full command: {}".format(dct_config_data)
        notif = ct.STATUS_TYPE.STATUS_EXCEPTION
        notif_code=ct.NOTIFICATION_CODES.PIPELINE_FAILED
      #endif success or not
    #endif good request or not
    
    self.P(msg + ". " + info, color='r' if notif == ct.STATUS_TYPE.STATUS_EXCEPTION else 'g')
    self._create_notification(
      notif=notif, 
      notif_code=notif_code,
      msg=msg,
      info=info,
      stream_name=pipeline_name,
      signature=signature,
      instance_id=instance_id,
      session_id=session_id,
      initiator_id=initiator_id,
      displayed=True,
    )
    if notif == ct.STATUS_TYPE.STATUS_EXCEPTION:
      return None
    else:
      return (pipeline_name, signature, instance_id)
    
    


  def batch_update_pipeline_instance(self, lst_configs, initiator_id, session_id):
    """
    Will update a list of pipelines. Each pipeline is updated with `update_pipeline_instance`
    

    Parameters
    ----------
    lst_configs : list
        the list of dicts
        
    Example
    -------
    
    ```json
    {
      "ACTION"  : "BATCH_UPDATE_PIPELINE_INSTANCE",
      "PAYLOAD" : [
        {
          "NAME"            : "test_for_batch_1",
          "SIGNATURE"       : "REST_CUSTOM_EXEC_01",
          "INSTANCE_ID"     : "RECE01_DFLT",
          "INSTANCE_CONFIG" : {
            "key11" : "value100",
            "key12" : "value200"
          }
        },
        {
          "NAME"            : "test_for_batch_2",
          "SIGNATURE"       : "REST_CUSTOM_EXEC_01",
          "INSTANCE_ID"     : "RECE01_DFLT",
          "INSTANCE_CONFIG" : {
            "key21" : "value1",
            "key22" : "value2"
          }
        }
      ]
    }
    ```
    
    TODO: must retrieve the address and pass it to each update_pipeline_instance call
    
    """
    results = []
    if not (isinstance(lst_configs, list) and len(lst_configs) >=1):
      msg = "Received invalid '{}' command. Command should contain a list of pipeline updates".format(
        ct.PAYLOAD_CT.COMMANDS.BATCH_UPDATE_PIPELINE_INSTANCE,
      )
      notif = ct.STATUS_TYPE.STATUS_EXCEPTION
      info = "Full received payload: {}".format(lst_configs)
    else:
      instances = [
        "  - {}:{}:{}".format(x.get(ct.PAYLOAD_DATA.NAME), x.get(ct.PLUGIN_INFO.SIGNATURE), x.get(ct.PLUGIN_INFO.INSTANCE_ID)) 
        for x in lst_configs
      ]
      self.P("Updating {} pipelines:\n{}".format(len(lst_configs), len(instances))) # TODO: delete 2nd len after testing
      for dct_config in lst_configs:
        res = self.update_pipeline_instance(
          dct_config, 
          session_id=session_id, initiator_id=initiator_id,
        )
        if res is not None:
          results.append(res)
      #endfor
      msg = "Successfully updated {} pipelines with `batch_update_pipeline_instance`".format(len(lst_configs))
      info = " Pipelines:\n  - " +"\n  - ".join([
        str(res)
        for res in results
      ])
      notif = ct.STATUS_TYPE.STATUS_NORMAL
    #endif valid command
    
    self.P(msg + ". " + info, color='r' if notif == ct.STATUS_TYPE.STATUS_EXCEPTION else 'g')
    self._create_notification(
      notif=notif, 
      msg=msg,
      info=info,
      displayed=True,
      BATCH_UPDATE_PIPELINE_INSTANCE=results,
      initiator_id=initiator_id,
      session_id=session_id,
    )
    return
  
  
  
  def pipeline_command(self, dct_config_data, initiator_id, session_id, **kwargs):
    pipeline_name = dct_config_data.get(ct.CONFIG_STREAM.NAME) 
    command = dct_config_data.get(ct.CONFIG_STREAM.PIPELINE_COMMAND)
    sender_addr = dct_config_data.get(ct.COMMS.COMM_RECV_MESSAGE.K_SENDER_ADDR, None)
    info = ""
    if pipeline_name not in self.dct_config_streams:
      msg = "Received invalid data '{}...' for commanding pipeline '{}'. Pipeline does not exist".format(
        str(dct_config_data)[:10], pipeline_name
      )
      notif = ct.STATUS_TYPE.STATUS_EXCEPTION
    else:      
      dct_pipeline = self.dct_config_streams[pipeline_name]
      
      if session_id is None: 
        session_id = dct_pipeline.get(ct.PAYLOAD_DATA.SESSION_ID)
        
      if initiator_id is None:
        initiator_id = dct_pipeline.get(ct.PAYLOAD_DATA.INITIATOR_ID)    
      else:
        # update modified-by fields
        dct_pipeline[ct.CONFIG_STREAM.K_MODIFIED_BY_ID] = initiator_id
        dct_pipeline[ct.CONFIG_STREAM.K_MODIFIED_BY_ADDR] = sender_addr
      #endif  
      
        
      dct_pipeline[ct.CONFIG_STREAM.PIPELINE_COMMAND] = command
      notif = ct.STATUS_TYPE.STATUS_NORMAL      
      self._save_stream_config(dct_pipeline) # save with null command
      msg = "Prepared pipeline command for pipeline '{}'".format(pipeline_name)
    #endif pipeline exists or not

    self.P(msg + ". " + info, color='r' if notif == ct.STATUS_TYPE.STATUS_EXCEPTION else 'g')
    self._create_notification(
      notif=notif, 
      msg=msg,
      info=info,
      stream_name=pipeline_name,
      session_id=session_id,
      initiator_id=initiator_id,
      displayed=True,
    )
    return     
  
  
  def delete_config_stream(self, stream_name, initiator_id, session_id, verbose=1):
    if stream_name not in self.dct_config_streams:
      msg = "Attempted to delete pipeline '{}' that does not exist. Current pipelines: {}".format(
        stream_name, list(self.dct_config_streams.keys())
      )
      self.P(msg, color='r')
      self._create_notification(
        notif=ct.STATUS_TYPE.STATUS_ABNORMAL_FUNCTIONING, 
        msg=msg,
        stream_name=stream_name,
        displayed=True,
      )
      return
    elif stream_name.lower() == self.admin_pipeline_name.lower():
      msg = "WARNING: Attempted detected to delete administration pipeline"
      self.P(msg, color='r')
      self._create_notification(
        notif=ct.STATUS_TYPE.STATUS_NORMAL, 
        notif_code=ct.NOTIFICATION_CODES.PIPELINE_ARCHIVE_FAILED,
        msg=msg,
      )
      return      
    #endif
    if session_id is None:
      session_id = self.dct_config_streams[stream_name].get(ct.PAYLOAD_DATA.SESSION_ID)
    if initiator_id is None:
      initiator_id = self.dct_config_streams[stream_name].get(ct.PAYLOAD_DATA.INITIATOR_ID)

    self._delete_stream_config(stream_name)
    self.dct_config_streams.pop(stream_name)
    
    msg = "Successfully deleted pipeline {}".format(stream_name)    
    if verbose >= 1:
      self.P(msg, color='g')
    
    self._create_notification(
      notif=ct.STATUS_TYPE.STATUS_NORMAL, 
      msg=msg,
      stream_name=stream_name,
      session_id=session_id,
      initiator_id=initiator_id,
      displayed=verbose >= 1,
    )
    #endif
    return
  

  def archive_single_stream(self, stream_name, initiator_id, session_id, verbose=1):    
    if not isinstance(stream_name, str):
      msg = "Attempt to archive a pipeline with no/wrong pipeline name/information: {}.".format(stream_name)
      self.P(msg, color='r')
      self._create_notification(
        notif=ct.STATUS_TYPE.STATUS_ABNORMAL_FUNCTIONING, 
        notif_code=ct.NOTIFICATION_CODES.PIPELINE_ARCHIVE_FAILED,
        session_id=session_id,
        initiator_id=initiator_id,
        msg=msg, 
        stream_name=stream_name,
        ct=ct,
        displayed=True
      )      
      return
    elif stream_name not in self.dct_config_streams:
      msg = "Attempted to archive pipeline '{}' that does not exist. Current pipelines: {}".format(
        stream_name, list(self.dct_config_streams.keys())
      )
      self.P(msg, color='r')
      self._create_notification(
        notif=ct.STATUS_TYPE.STATUS_ABNORMAL_FUNCTIONING, 
        notif_code=ct.NOTIFICATION_CODES.PIPELINE_ARCHIVE_FAILED,
        session_id=session_id,
        initiator_id=initiator_id,
        msg=msg,
        stream_name=stream_name,
        ct=ct,
        displayed=True,        
      )
      return
    elif stream_name.lower() == self.admin_pipeline_name.lower():
      msg = "WARNING: Attempted detected to archive administration pipeline"
      self.P(msg, color='r')
      self._create_notification(
        notif=ct.STATUS_TYPE.STATUS_NORMAL, 
        notif_code=ct.NOTIFICATION_CODES.PIPELINE_ARCHIVE_FAILED,
        session_id=session_id,
        initiator_id=initiator_id,
        msg=msg,
      )
      return      

    #endif bad stuff
    if session_id is None:
      session_id = self.dct_config_streams[stream_name].get(ct.PAYLOAD_DATA.SESSION_ID)
    if initiator_id is None:
      initiator_id = self.dct_config_streams[stream_name].get(ct.PAYLOAD_DATA.INITIATOR_ID)
    
    self.P("Archiving pipeline '{}' from '{}', sessid: {}...".format(stream_name, initiator_id, session_id))

    self._save_stream_config(
      self.dct_config_streams[stream_name],
      subfolder_path=os.path.join(self._folder_streams_configs, 'archived_streams'),
      prefix_fn=self.log.now_str(nice_print=False, short=True)
    )

    self.delete_config_stream(
      stream_name=stream_name, 
      session_id=session_id, initiator_id=initiator_id,
      verbose=0,
    )

    msg = "Successfully archived (and deleted) pipeline '{}' (currently {} active pipelines)".format(stream_name, len(self.dct_config_streams))
    self.P(msg)
      
    self._create_notification(
      notif=ct.STATUS_TYPE.STATUS_NORMAL, 
      notif_code=ct.NOTIFICATION_CODES.PIPELINE_ARCHIVE_OK,
      msg=msg,
      stream_name=stream_name,
      session_id=session_id,
      initiator_id=initiator_id,
      ct=ct,
      displayed=True,
    )
    return

  

  
  def archive_streams(self, stream_names, initiator_id, session_id):
    for stream_name in stream_names:
      self.archive_single_stream(stream_name, initiator_id=initiator_id, session_id=session_id)
    return

  
  def archive_all_streams(self, initiator_id, session_id):
    lst_streams = list(self.dct_config_streams.keys())
    self.P("Proceeding to ARCHIVING of {}".format(lst_streams))
    self.archive_streams(lst_streams, initiator_id=initiator_id, session_id=session_id)
    return


  def delete_streams(self, stream_names, initiator_id, session_id):
    for stream_name in stream_names:
      self.delete_config_stream(stream_name, initiator_id=initiator_id, session_id=session_id)
    return

  
  def delete_all_streams(self, initiator_id, session_id):
    lst_streams = list(self.dct_config_streams.keys())
    self.P("Proceeding to DELETION of {}".format(lst_streams))
    self.delete_streams(lst_streams, initiator_id=initiator_id, session_id=session_id)
    return


from naeural_core.utils.config_utils import get_config_from_code

_CONFIG  = {
  "TEST" : 1,
  "TEST2" : {
     "1" : 2,
     "3" : [1,2,3],
     "5" : {"A": 1, "B": 0}
     
    }
  }




if __name__ == '__main__':

  print(get_config_from_code(__file__))


"""
Example pipeline:

{
  "NAME" : "IoT-Example",
  "TYPE" : "IoTQueueListener",
  
  "PATH_FILTER" : [
    None, None, 
    ["IOT_PLUGIN_DEMO_1", "IOT_PLUGIN_DEMO_2"], 
    None
  ],
  "MESSAGE_FILTER" : {},
  
  "PLUGINS" : [
    {
      "SIGNATURE" : "IOT_PLUGIN_DEMO_1",
      "INSTANCES" : [
        {
          "INSTANCE_ID" : "IOT_PLUGIN_DEMO_1_INSTANCE_1",
          "USE_TYPE" : "A",
          "PROCESS_DELAY" : 0.1
        }
      ]
    },
    {
      "SIGNATURE" : "IOT_PLUGIN_DEMO_1",
      "INSTANCES" : [
        {
          "INSTANCE_ID" : "IOT_PLUGIN_DEMO_1_INSTANCE_2",
          "USE_TYPE" : "B",
          "PROCESS_DELAY" : 0.1
        }
      ]
    },
    {
      "SIGNATURE" : "IOT_PLUGIN_DEMO_1",
      "INSTANCES" : [
        {
          "INSTANCE_ID" : "IOT_PLUGIN_DEMO_1_INSTANCE_3",
          "USE_TYPE" : "C",
          "PROCESS_DELAY" : 0.1
        }
      ]
    }        
  ]
}


IOT_PLUGIN_DEMO_1:
  process:
    if datainput['USE_TYPE'] != MYTYPE:
      return
    do_the_processing(datainput)

"""


from naeural_core.data.base.base_iot_queue_listener import \
    BaseIoTQueueListenerDataCapture

_CONFIG = {
  **BaseIoTQueueListenerDataCapture.CONFIG,

  'VALIDATION_RULES': {
    **BaseIoTQueueListenerDataCapture.CONFIG['VALIDATION_RULES'],
  },
}


class IoTQueueListenerDataCapture(BaseIoTQueueListenerDataCapture):
  CONFIG = _CONFIG

  def __init__(self, **kwargs):
    super(IoTQueueListenerDataCapture, self).__init__(**kwargs)
    return

  def _filter_message(self, unfiltered_message):
    filtered_message = unfiltered_message
    return filtered_message

  def _parse_message(self, filtered_message):
    return filtered_message



import torch as th
import numpy as np
from shapely import Polygon


def intersect_over_union(th_y, th_y_hat):
  """
  :param y:
    Target tensor with size (bs, 4)
  :param y_hat:
    Predicted tensor with size (bs, 4)
  :return:
    IOU with size (bs)
  """
  dev1 = th_y.device
  dev2 = th_y_hat.device
  assert dev1 == dev2, 'Tensors should be placed on the same device!'

  th_xA = th.max(th_y[:, 0], th_y_hat[:, 0])
  th_yA = th.max(th_y[:, 1], th_y_hat[:, 1])
  th_xB = th.min(th_y[:, 2], th_y_hat[:, 2])
  th_yB = th.min(th_y[:, 3], th_y_hat[:, 3])
  th_inter_area = th.max(th.zeros(th_xA.shape).to(dev1), th_xB - th_xA + 0.01) * \
      th.max(th.zeros(th_yA.shape).to(dev1), th_yB - th_yA + 0.01)
  th_boxA_area = (th_y[:, 2] - th_y[:, 0] + 0.01) * (th_y[:, 3] - th_y[:, 1] + 0.01)
  th_boxB_area = (th_y_hat[:, 2] - th_y_hat[:, 0] + 0.01) * (th_y_hat[:, 3] - th_y_hat[:, 1] + 0.01)
  th_iou = th_inter_area / (th_boxA_area + th_boxB_area - th_inter_area)
  return th_iou


def intersect_over_union_polygons(th_y, th_y_hat):
  """
  :param y:
    Target tensor with size (bs, 4, 2)
  :param y_hat:
    Predicted tensor with size (bs, 4, 2)
  :return:
    IOU with size (bs)
  """
  dev1 = th_y.device
  dev2 = th_y_hat.device
  assert dev1 == dev2, 'Tensors should be placed on the same device!'

  ious = []
  for poly1, poly2 in zip(th_y, th_y_hat):
    # Define each polygon
    polygon1_shape = Polygon(poly1)
    polygon2_shape = Polygon(poly2)

    # Calculate intersection and union, and tne IOU
    polygon_intersection = polygon1_shape.intersection(polygon2_shape).area
    polygon_union = polygon1_shape.area + polygon2_shape.area - polygon_intersection
    ious.append(polygon_intersection / polygon_union)

  return th.tensor(ious)


def intersect_over_gtarea(y, y_hat):
  """
  :param y:
    Target tensor with size (bs, 4)
  :param y_hat:
    Predicted tensor with size (bs, 4)
  :return:
    IOU with size (bs)
  """

  xA = th.max(y[:, 0], y_hat[:, 0])
  yA = th.max(y[:, 1], y_hat[:, 1])
  xB = th.min(y[:, 2], y_hat[:, 2])
  yB = th.min(y[:, 3], y_hat[:, 3])
  interArea = th.max(th.zeros(xA.shape), xB - xA + 0.01) * th.max(th.zeros(yA.shape), yB - yA + 0.01)
  boxAArea = (y[:, 2] - y[:, 0] + 0.01) * (y[:, 3] - y[:, 1] + 0.01)
  ioa = interArea / boxAArea
  return ioa


def neighbour_accuracy(y, y_pred, neighbour_distance=1, neighbour_multiplier=1):
  assert neighbour_distance > 0, "`neighbour_distance` must be positive"
  exact_preds = np.sum(np.abs(y - y_pred) == 0)
  neighbour_preds = np.sum((0 < np.abs(y - y_pred)) & (np.abs(y - y_pred) <= neighbour_distance)) * neighbour_multiplier
  return (exact_preds + neighbour_preds) / y.shape[0]



from naeural_core.serving.ai_engines import AI_ENGINES

def _handle_ai_engine(ai_engine):
  model_instance_id = 'DEFAULT'
  if isinstance(ai_engine, (tuple, list)):
    assert len(ai_engine) == 2
    ai_engine, model_instance_id = ai_engine

  if '?' in ai_engine:
    ai_engine, model_instance_id = ai_engine.split('?')

  return {
    'AI_ENGINE' : ai_engine,
    'MODEL_INSTANCE_ID' : model_instance_id
  }

def get_serving_process_given_ai_engine(ai_engine):
  handle = _handle_ai_engine(ai_engine)
  ai_engine = handle['AI_ENGINE'].lower()
  instance_id = handle['MODEL_INSTANCE_ID']

  config_ai_engine = AI_ENGINES.get(ai_engine, {'SERVING_PROCESS': ai_engine})
  serving_process = config_ai_engine['SERVING_PROCESS']
  if config_ai_engine.get('REQUIRES_INSTANCE', False):
    # in this case we must code the serving process as CLASS, NAME particularly
    # for custom downloadable models
    # TODO: however the support seems to be lost downstream!
    serving_process = serving_process, instance_id
  return serving_process

def get_ai_engine_given_serving_process(serving_process, params={}):
  model_instance_id = None
  if isinstance(serving_process, (tuple, list)):
    assert len(serving_process) == 2
    serving_process, model_instance_id = serving_process

  for ai_engine, dct in AI_ENGINES.items():
    crt_params = dct.get('PARAMS', {})
    crt_serving_process = dct['SERVING_PROCESS']
    if serving_process == crt_serving_process and params == crt_params:
      if model_instance_id is None:
        return ai_engine
      else:
        return ai_engine, model_instance_id
  #endfor

  if model_instance_id is None:
    return serving_process
  else:
    return serving_process, model_instance_id

def get_params_given_ai_engine(ai_engine):
  handle = _handle_ai_engine(ai_engine)
  ai_engine = handle['AI_ENGINE']
  config_ai_engine = AI_ENGINES.get(ai_engine, {'SERVING_PROCESS': ai_engine})
  other_params = config_ai_engine.get('PARAMS', {})
  return other_params

def get_params_given_serving_process(serving_process):
  # this is tricky as multiple AI_ENGINES could share the same serving process
  # thus this function should work ONLY if there is just one serving_process
  model_instance_id = None
  if isinstance(serving_process, (tuple, list)):
    assert len(serving_process) == 2
    serving_process, model_instance_id = serving_process
  lst_found = []
  for ai_engine, dct in AI_ENGINES.items():
    crt_serving_process = dct['SERVING_PROCESS']
    if crt_serving_process == serving_process:
      lst_found.append(dct)
  if len(lst_found) == 1:
    return lst_found[0].get('PARAMS', {})
  else:
    # either nothing found or multiple found
    if len(lst_found) > 1:
      raise ValueError("Multiple AI Engines use '{}' - cannot retrieve the params!".format(serving_process))
    else:
      raise ValueError("NO AI Engines use '{}' - cannot retrieve the params!".format(serving_process))
  return
      
  



import json
from tqdm import tqdm
from time import sleep
from naeural_core import Logger
from datetime import datetime
from comm.comm_plugins.paho_comm import PahoCommThread

CONFIG_SENDER = {
  'DEVICE_ID'             : '12345',
  'HOST'                  : 'HOSTNAME',
  'PORT'                  : 1883,
  'USER'                  : 'USERNAME',
  'PASS'                  : 'PASSWORD',
  'PATH'                  : 'PATH',
  'TOPIC_SEND'            : 'config_send_recv',
  'TOPIC_RECV'            : 'not_used',
  'TOPIC_DEVICE_SPECIFIC' : False,
  'QOS'                   : 1
}
CONFIG_RECEIVER = {
  'DEVICE_ID'             : '54321',
  'HOST'                  : 'HOSTNAME',
  'PORT'                  : 1883,
  'USER'                  : 'USERNAME',
  'PASS'                  : 'PASSWORD',
  'PATH'                  : 'PATH',
  'TOPIC_SEND'            : 'not_used',
  'TOPIC_RECV'            : 'config_send_recv',
  'TOPIC_DEVICE_SPECIFIC' : False,
  'QOS'                   : 1
}

def test_connection(comm):
  log.p('Running `test_connection`')
  is_ok = False
  t_sleep = 0.1
  nr_retry = 5
  has_conn = False
  i = 0
  while i < nr_retry:    
    has_conn = comm.has_send_conn and comm.has_recv_conn
    if has_conn:
      break
    i+= 1
    comm.dump_log()
    sleep(t_sleep)
  if not has_conn: 
    log.p('Test failed!', color='r')
  else:
    is_ok = True
    log.p('Communication passed!', color='g')
  return is_ok

def test_send_receive(comm_send, comm_recv, n_messages=200):
  log.p('Running `test_send_receive`')
  t_sleep = 0.1
  l_recv = []
  l_send = [{'ID': i, 'TEXT': 'TEST_{}'.format(i)} for i in range(n_messages)]
  l_send_test= l_send + 10 * [None] #some dummy messages that will not be send
  for msg in tqdm(l_send_test):
    if msg:
      comm_send.send(msg)
    sleep(t_sleep)
    l_recv+= comm_recv.get_messages()
  #endfor
  
  if not l_send == l_recv:
    log.p('Test failed! {} sent vs {} received'.format(len(l_send), len(l_recv)), color='r')
  else:
    log.p('Test passed!', color='g')
  return l_send, l_recv


if __name__ == '__main__':
  cfg_file = 'main_config.txt'
  log = Logger(
    lib_name='PAHO_TEST', 
    config_file=cfg_file, 
    max_lines=1000, 
    TF_KERAS=False
    )
  
  #start comm  
  comm_send = PahoCommThread(
    log=log,
    config_sender=CONFIG_SENDER,
    config_receiver=None
    )
  comm_send.start()
  
  comm_recv = PahoCommThread(
    log=log,
    config_sender=CONFIG_RECEIVER,
    config_receiver=None
    )
  comm_recv.start()
  
  #allow some time to setup communication with the server
  sec_sleep = 5
  start = datetime.now()
  while (datetime.now() - start).seconds <= sec_sleep:
    comm_send.dump_log()
    comm_recv.dump_log()
  
  send_ok = test_connection(comm_send)
  recv_ok = test_connection(comm_recv)
  if send_ok and recv_ok:
    l_send, l_recv = test_send_receive(comm_send, comm_recv)
  # test_send_receive_2(comm)
    


import asyncio
import uuid
import multiprocessing
from multiprocessing.managers import SyncManager

# The following need to be global due to the multiprocessing
# method being set to spawn.
class CommsServerManager(SyncManager):
  pass
class CommsClientManager(SyncManager):
  pass

server_queue = None
def get_server_queue():
  global server_queue
  if server_queue is None:
    server_queue = multiprocessing.Queue()
  return server_queue

client_queue = None
def get_client_queue():
  global client_queue
  if client_queue is None:
    client_queue = multiprocessing.Queue()
  return client_queue


def get_server_manager(auth):
    CommsServerManager.register('get_server_queue', callable=get_server_queue)
    CommsServerManager.register('get_client_queue', callable=get_client_queue)
    manager = CommsServerManager(
      address=('127.0.0.1', 0),
      authkey=auth
    )
    manager.start()
    return manager

def get_client_manager(port, auth):
    CommsClientManager.register('get_server_queue')
    CommsClientManager.register('get_client_queue')
    manager = CommsClientManager(
      address=('127.0.0.1', port),
      authkey=auth
    )
    manager.connect()
    return manager

class UvicornPluginComms:
  """
  Communicator for Uvicorn/FastAPI with an instance of the fastapi plugin.
  Since this is meant to be used on the web server side, it's methods
  are asynchronous and will use the FastAPI even loop to deliver
  messages.

  FastAPI endpoints can call the call_plugin method to deliver and receive
  requests from the associated business plugin.
  """
  def __init__(self, port, auth):
    """
    Initializer for the UvicornPluginComms.

    Parameters
    ----------
    port: int, port value used for commuication with the business plugin
    auth: str, string value used for authenticating and establishing
      communications with the business plugin.
    Returns
    -------
    None
    """
    self.manager = get_client_manager(
      port=port, auth=auth
    )
    self._commands = {}
    self._reads_messages = False
    return

  async def call_plugin(self, *request):
    """
    Calls the business plugin with the supplied method and arguments.

    Parameters
    ----------
    First parameter is the method name of the business plugin that should
    be invoked to process this request. Subsequent parameters are the
    arguments of this method.

    Returns
    -------
    Any - the reply from the business plugin for this request

    Example
    -------
    await comms.call_plugin('foo', 'bar', 'baz') will call the
    business plugin method 'foo' with arguments 'bar' and 'baz'.
    """
    # Generate a uuid for the current message and an event that can be
    # used to signal that the request has been completed.
    event = asyncio.Event()
    ee_uuid = uuid.uuid4()
    ee_uuid = str(ee_uuid)[:8]

    # Add an entry with the current uuid, event and a placeholder for
    # the reply ('value') so that the communication even loop can
    # signal the work being completed.
    self._commands[ee_uuid] = {
      'value' : None,
      'event' : event
    }

    # Send the request to the business plugin.
    self.manager.get_server_queue().put({
      'id' : ee_uuid,
      'value' : request
    })

    if not self._reads_messages:
      # Insert the event processing task into the current even loop if we've
      # not already done that. This needs to be done after the event loop
      # has already been started, which is why we need to do it here.
      self._reads_messages = True
      loop = asyncio.get_running_loop()
      asyncio.ensure_future(self._read_from_plugin(), loop=loop)
    #endif maybe start even loop

    # Yield until we've received a reply back from the business plugin.
    await event.wait()

    # We now have a reply, retrieve the reply and cleanup our entry from the
    # commands dict.
    response = self._commands[ee_uuid]['value']
    del self._commands[ee_uuid]
    return response

  async def _read_from_plugin(self):
    """
    Starts the task of continuously reading messages from the business plugin.
    When a message is recieved, the corresponding request created by
    call_plugin will be signaled to resume and read the reply from the _commands
    dict.

    This is internal to the functioning of the communicator and should NOT be
    directly called by users.

    Parameters
    ----------
    None

    Returns
    -------
    None
    """
    while True:
      client_queue = self.manager.get_client_queue()
      while not client_queue.empty():
        message = client_queue.get()
        ee_uuid = message['id']
        record = self._commands[ee_uuid]
        record['value'] = message['value']
        record['event'].set()
      #endwhile read all messages from the queue

      # We've done all the work for now. Yield and look again for messages in
      # 0.1 seconds
      await asyncio.sleep(0.001)
    #endwhile communicator loop
    return




__VER__ = '1.2.1'


class _NetworkProcessorMixin:
  
  @staticmethod
  def payload_handler(signature="DEFAULT"):
    if not isinstance(signature, str):
      signature = "DEFAULT"
    signature = signature.upper()
    def decorator(f):
      f.__payload_signature__ = signature
      return f
    return decorator
  

  def network_processor_init(self):
    self.__non_dicts = 0
    self.__handlers = {}
    # we get all the functions that start with on_payload_
    for name in dir(self):
      if callable(getattr(self, name)):
        func = getattr(self, name)
        if name.startswith("on_payload_"):
          signature = name.replace("on_payload_", "").upper()
          self.__handlers[signature] = getattr(self, name)
        # end if we have a on_payload_<signature>
        if hasattr(func, "__payload_signature__"):
          signature = func.__payload_signature__.upper()
          if signature == "DEFAULT":
            signature = self._signature.upper()
          self.__handlers[signature] = getattr(self, name)
        # end if we have a signature
      # end if callable
    # end for each name in dir
        
    if len(self.__handlers) == 0:
      self.P("No payload handlers found", color="red")
    else:
      self.P("Payload handlers found for: {}".format(list(self.__handlers.keys())), color="green")
    self._network_processor_initialized = True
    return
  
  def __network_processor_maybe_check_initialized(self):
    if not hasattr(self, "_network_processor_initialized") or not self._network_processor_initialized:
      msg = "NetworkProcessorPlugin not initialized probably due to missing super().on_init() in child class"
      self.P(msg, color="red")
      raise ValueError(msg)
      return False
    return True


  def get_instance_path(self):
    return [self.ee_addr, self._stream_id, self._signature, self.cfg_instance_id]
  
  
  def __network_processor_maybe_process_received(self):
    datas = self.dataapi_struct_datas(full=False, as_list=True)    
    assert isinstance(datas, list), f"Expected list but got {type(datas)}"
    if len(datas) > 0:
      for data in datas:
        if not isinstance(data, dict):
          self.__non_dicts += 1
          if self.cfg_full_debug_payloads:
            self.P(f"Received non dict payload: {data} from {datas}", color="red")           
          continue
        varified = False
        verify_msg = None
        try:
          verify_results = self.bc.verify(
            dct_data=data, 
            str_signature=None, sender_address=None,
            return_full_info=True,
          )
          verified = verify_results.valid
          verify_msg = verify_results.message
        except Exception as e:
          self.P(f"{e}: {data}", color="red")
          continue
        if not verified:
          self.P(
            f"Payload signature verification FAILED with '{verify_msg}': {data}", 
            color="red"
          )
          continue
        payload_path = data.get(self.const.PAYLOAD_DATA.EE_PAYLOAD_PATH, [None, None, None, None])        
        is_self = payload_path == self.get_instance_path()
        if is_self and not self.cfg_accept_self:
          continue
        signature = payload_path[2]
        sender = data.get(self.const.PAYLOAD_DATA.EE_SENDER, None)
        signature = signature.upper()
        if signature in self.__handlers:
          if self.cfg_full_debug_payloads:
            self.P(f"RECV-{signature} <{sender}>: {payload_path}")
          self.__handlers[signature](data)
        else:
          if self.cfg_full_debug_payloads:
            self.P(f"RECV-UNKNOWN <{sender}>: {payload_path}")
        # end if we have handlers
      # for each data observation in dct datas     
    # end if we have payloads
    return


  def network_processor_loop(self):
    """
    This method must be protected while the child plugins should have normal `process`
    """
    self.__network_processor_maybe_check_initialized()
    self.__network_processor_maybe_process_received()  
    return



from naeural_core import Logger
import json

if __name__ == '__main__':
  
  
  d2 = {
    1: [1,2,3],
    2: {
      22: 22,
      33 : [33,33],
      44 : [{55:55}]
      }
    }
  
  l = Logger('DCT', base_folder='.', app_folder='_local_cache')
  
  with open('core/xperimental/dict/gql_resp.json', 'rt') as f:
    d = json.load(f)
    
  l.dict_pretty_format(d, display=True)
  l.dict_pretty_format(d2, display=True)


import tensorflow as tf
import tensorflow_hub as hub
import os
import numpy as np
import traceback

from naeural_core import Logger

POSE_POINTS = {
  'nose': 0,
  'left_eye': 1,
  'right_eye': 2,
  'left_ear': 3,
  'right_ear': 4,
  'left_shoulder': 5,
  'right_shoulder': 6,
  'left_elbow': 7,
  'right_elbow': 8,
  'left_wrist': 9,
  'right_wrist': 10,
  'left_hip': 11,
  'right_hip': 12,
  'left_knee': 13,
  'right_knee': 14,
  'left_ankle': 15,
  'right_ankle': 16
  }

class ModelWrapper(tf.Module):
  def __init__(self, model):
    self.model = model
    
  @tf.function(input_signature=(tf.TensorSpec(shape=(1,256, 256, 3), dtype=tf.int32),))
  def predict(self, inputs):
    return model(inputs)

if __name__ == '__main__':
  l = Logger("MVT", base_folder='.', app_folder='_cache')
  img = np.random.randint(0, 255, (1, 256, 256, 3), dtype='uint8')
  tf_img = tf.constant(img, dtype='int32')
  LOAD = False
  USE_SIGN = True
  fn = os.path.join(l.get_models_folder(), "movenet")
  if LOAD:
    l.P("Loading module", color='g')
    module = hub.load("https://tfhub.dev/google/movenet/singlepose/thunder/4")
    model = module.signatures['serving_default']
    pred = model(tf_img)
    try:
      l.P("Saving module to {}".format(fn))
      tf.saved_model.save(module, fn)
    except:
      l.P("Exception: {}".format(traceback.format_exc()))
  else:
    try:
      l.P("loading module to {}".format(fn))
      module = tf.saved_model.load(fn)
      model = module.inference_fn
      pred = model(tf_img).numpy()
      l.P("Result: {}".format(pred.shape), color='g')
    except:
      l.P("Exception: {}".format(traceback.format_exc()))
      

  


from cryptography.hazmat.primitives import serialization, hashes
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.backends import default_backend
import base64
import hashlib
import json

# Generate keys
private_key = ec.generate_private_key(ec.SECP256K1(), default_backend())
public_key = private_key.public_key()

# Serialize public key in X9.62 compressed point format
public_bytes = public_key.public_bytes(
    encoding=serialization.Encoding.X962,
    format=serialization.PublicFormat.CompressedPoint
)

# Hash the message
# dct_to_send = { '9': 9, '2': 2, '3': 3, '10': { '2': 2, '100': 100, '1': 1 } }
str_to_send = """{"EE_FORMATTER":"cavi2","EE_MESSAGE_ID":"a512448e-8f9e-4e9b-a3b3-d8071bbaee3f","EE_PAYLOAD_PATH":["dev-1","Cam-Radu-1",null,null],"SB_IMPLEMENTATION":"cavi2","category":"","data":{"identifiers":{},"img":{"height":null,"id":null,"width":null},"specificValue":{},"time":null,"value":{}},"demoMode":false,"messageID":"a512448e-8f9e-4e9b-a3b3-d8071bbaee3f","metadata":{"displayed":true,"ee_message_seq":33501,"ee_timezone":"UTC+3","ee_tz":"Europe/Bucharest","error_code":null,"info":"EE v3.28.131, Lib v9.8.62, info text: None","initiator_id":"cavi2-local-radu","module":"VideoStreamDataCapture","notification":"Video DCT 'Cam-Radu-1' successfully connected. Overall 8231 reconnects.","notification_type":"NORMAL","sbCurrentMessage":"a512448e-8f9e-4e9b-a3b3-d8071bbaee3f","sbTotalMessages":33501,"sb_event_type":"NOTIFICATION","sb_id":"dev-1","session_id":null,"stream_name":"Cam-Radu-1","timestamp":"2023-10-19 14:49:03.680131","video_stream_info":{"buffersize":0,"current_interval":null,"fps":23,"frame_count":429,"frame_current":3513330,"frame_h":1080,"frame_w":1920}},"sender":{"hostId":"dev-1","id":"edge-node","instanceId":"edge-node-v3.28.131"},"time":{"deviceTime":"","hostTime":"2023-10-19 14:49:03.982422","internetTime":""},"type":"notification","version":"3.28.131"}"""
dct_to_send = {
  "EE_FORMATTER": "cavi2",
  "EE_MESSAGE_ID": "a512448e-8f9e-4e9b-a3b3-d8071bbaee3f",
  "EE_PAYLOAD_PATH": [
    "dev-1",
    "Cam-Radu-1",
    None,
    None
  ],
  "SB_IMPLEMENTATION": "cavi2",
  "category": "",
  "data": {
    "identifiers": {},
    "img": {
      "height": None,
      "id": None,
      "width": None
    },
    "specificValue": {},
    "time": None,
    "value": {}
  },
  "demoMode": False,
  "messageID": "a512448e-8f9e-4e9b-a3b3-d8071bbaee3f",
  "metadata": {
    "displayed": True,
    "ee_message_seq": 33501,
    "ee_timezone": "UTC+3",
    "ee_tz": "Europe/Bucharest",
    "error_code": None,
    "info": "EE v3.28.131, Lib v9.8.62, info text: None",
    "initiator_id": "cavi2-local-radu",
    "module": "VideoStreamDataCapture",
    "notification": "Video DCT 'Cam-Radu-1' successfully connected. Overall 8231 reconnects.",
    "notification_type": "NORMAL",
    "sbCurrentMessage": "a512448e-8f9e-4e9b-a3b3-d8071bbaee3f",
    "sbTotalMessages": 33501,
    "sb_event_type": "NOTIFICATION",
    "sb_id": "dev-1",
    "session_id": None,
    "stream_name": "Cam-Radu-1",
    "timestamp": "2023-10-19 14:49:03.680131",
    "video_stream_info": {
      "buffersize": 0,
      "current_interval": None,
      "fps": 23,
      "frame_count": 429,
      "frame_current": 3513330,
      "frame_h": 1080,
      "frame_w": 1920
    }
  },
  "sender": {
    "hostId": "dev-1",
    "id": "edge-node",
    "instanceId": "edge-node-v3.28.131"
  },
  "time": {
    "deviceTime": "",
    "hostTime": "2023-10-19 14:49:03.982422",
    "internetTime": ""
  },
  "type": "notification",
  "version": "3.28.131"
}
message = json.dumps(dct_to_send, sort_keys=True, separators=(',',':'))
bin_message = message.encode()
assert message == str_to_send
obj_hash = hashlib.sha256(bin_message)
str_hash = obj_hash.hexdigest()
print("Hash: "+ str_hash)
b_hash = obj_hash.digest()

# Sign the hash
signature_bytes = private_key.sign(b_hash, ec.ECDSA(hashes.SHA256()))

# Base64url encode the signature
signature_b64 = base64.urlsafe_b64encode(signature_bytes).decode()

pk_b64 = base64.urlsafe_b64encode(public_bytes).decode()

dct_msg = {
  **dct_to_send,
  'EE_SENDER': "0xai_" + pk_b64,
  'EE_SIGN': signature_b64,
  'EE_HASH': str_hash
}

str_msg = json.dumps(dct_msg) # no need for sorting
print("Message: '"+ str_msg + "'")






from naeural_core.serving.base.backends.ovino import OpenVINOModel
from naeural_core.utils.tracing.onnx.utils import create_from_torch
from xperimental.th_y8.generate import get_test_images
from xperimental.th_y8.utils import Y8, BackendType
from naeural_core.local_libraries.nn.th.utils import th_resize_with_pad

import torch as th
from ultralytics import YOLO

if __name__ == '__main__':

  device = th.device('cpu')
  imgs = get_test_images()
  imgs = th_resize_with_pad(
    img=imgs,
    h=640,
    w=640,
    device=device,
    normalize=True,
    half=False,
    return_original=False
  )[0].float()

  model = YOLO('yolov8n.pt').model
  config = {
    'foo' : 'bar'
  }
  model = Y8(
    (model, config),
    device,
    topk=False,
    backend_type=BackendType.ONNX
  )

  # Note no half support for CPU so half is always false here.
  path = '/tmp/yolo.onnx'
  create_from_torch(
    model, device, path, half=False,
    input_names=['input0'],
    output_names=['output0', 'output1'],
    args=imgs,
    aggressive_shape_inference=False,
    metadata=config)

  vmodel = OpenVINOModel()
  vmodel.load_model(model_path=path, half=True)
  imgs = imgs.half()
  print(vmodel.get_metadata())
  print(vmodel.get_input_dtype(0))
  o0, o1 = vmodel(imgs)
  print(o1.shape)

  o0, o1 = vmodel(imgs[:4])
  for i in range(o0.shape[0]):
    print('==============================')
    for j in range(o1[i].item()):
      print(o0[i, j, :].numpy())



from naeural_core.data.base.base_decentrai_connector import BaseDecentraiConnectorDataCapture

_CONFIG = {
  **BaseDecentraiConnectorDataCapture.CONFIG,

  'VALIDATION_RULES': {
    **BaseDecentraiConnectorDataCapture.CONFIG['VALIDATION_RULES'],
  },
}


class DecentraiConnectorDataCapture(BaseDecentraiConnectorDataCapture):
  CONFIG = _CONFIG

  def interpret_message(self, message):
    return message



"""
Run a rest API exposing the yolov5s object detection model
"""
import argparse
import io

import torch
from PIL import Image
from flask import Flask, request

app = Flask(__name__)

DETECTION_URL = "/v1/object-detection/yolov5s"


@app.route(DETECTION_URL, methods=["POST"])
def predict():
    if not request.method == "POST":
        return

    if request.files.get("image"):
        image_file = request.files["image"]
        image_bytes = image_file.read()

        img = Image.open(io.BytesIO(image_bytes))

        results = model(img, size=640)  # reduce size=320 for faster inference
        return results.pandas().xyxy[0].to_json(orient="records")


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Flask API exposing YOLOv5 model")
    parser.add_argument("--port", default=5000, type=int, help="port number")
    args = parser.parse_args()

    model = torch.hub.load("ultralytics/yolov5", "yolov5s", force_reload=True)  # force_reload to recache
    app.run(host="0.0.0.0", port=args.port)  # debug=True causes Restarting with stat




import json
import multiprocessing as mp
import os
import shutil
import sys
import traceback
import warnings

from uuid import uuid4

warnings.filterwarnings("ignore")

#local dependencies
from naeural_core import constants as ct
from naeural_core.main.orchestrator import Orchestrator
from naeural_core import Logger

from naeural_core.main.ver import __VER__

# TODO: change to `from ratio1 import`
from ratio1.utils import load_dotenv

MANDATORY_PACKAGES = {
  'torch'           : '2.0',
  'accelerate'      : '0.2',
  'transformers'    : '4.43',
  'tokenizers'      : '0.14',
}

def maybe_replace_txt(fn):
  result = fn
  if fn.endswith('.txt'):
    fn_dest = fn.replace('.txt', '.json')
    print("Found '{}' as base startup config file. Converting to '{}'...".format(fn, fn_dest), flush=True)
    # duplicate .txt to .json    
    shutil.copy(fn, fn_dest)
    # delete old .txt
    os.remove(fn)
    result = fn_dest
  return result

def running_with_hostname(config_file):
  result = None
  custom_ee_id = False
  ee_id = os.environ.get(ct.CONFIG_STARTUP_v2.K_EE_ID, '')[:ct.EE_ALIAS_MAX_SIZE]
  if len(ee_id) > 0:
    custom_ee_id = ee_id.upper().replace('X','') not in ['HOSTNAME', '']
    print("Found {} in env EE_ID: '{}' ".format("custom" if custom_ee_id else "default", ee_id), flush=True)    
  else:
    print("No EE_ID found in env", flush=True)
  is_hostname_config = False
  is_hostname_env = ee_id in ['HOSTNAME'] # if explicitly set to HOSTNAME in environment
  with open(config_file, 'r') as fh:
    config_data = json.load(fh)
    config_ee_id = config_data.get(ct.CONFIG_STARTUP_v2.K_EE_ID, '')[:ct.EE_ALIAS_MAX_SIZE]
    print("Found EE_ID in config: '{}'".format(config_ee_id), flush=True)
    str_simple = config_ee_id.upper().replace('X','')
    is_hostname_config = str_simple in ['HOSTNAME', ''] # if explicitly set to HOSTNAME in config or first run with no config
    if str_simple == 'HOSTNAME':
      custom_ee_id = False # if config is set to HOSTNAME, then we ignore the env EE_ID
  #endwith config
  if (is_hostname_env or is_hostname_config) and not custom_ee_id:
    default_uuid = str(uuid4())[:8]
    result = os.environ.get('HOSTNAME', default_uuid)
  return result
  
def get_id(log : Logger):
  config_box_id = log.config_data.get(ct.CONFIG_STARTUP_v2.K_EE_ID, '')[:ct.EE_ALIAS_MAX_SIZE]
  log.P("Found EE_ID '{}'".format(config_box_id))
  if config_box_id.upper().replace('X','').upper() in [None, '', 'HOSTNAME', 'E2DKR']:
    config_box_id_env = os.environ.get('EE_ID', '')
    if isinstance(config_box_id_env, str) and config_box_id_env.upper() not in ['', 'E2DKR','HOSTNAME']:
      config_box_id = config_box_id_env
    else:
      config_box_id = log.get_random_name(2)
      log.P("E2 is not manually configured nor from env. Assuming a random id '{}'".format(config_box_id), color='r')
    log.config_data[ct.CONFIG_STARTUP_v2.K_EE_ID] = config_box_id
    log.P("  Saving/updating config with new EE_ID '{}'...".format(config_box_id))
    log.update_config_values({ct.CONFIG_STARTUP_v2.K_EE_ID: config_box_id})
  #endif config not ok
  return config_box_id
  
  
def get_config(config_fn):  
  fn = None
  extensions = ['.json', '.txt',]
  for loc in ['.', ct.LOCAL_CACHE]:
    for ext in extensions:
      test_fn = os.path.join(loc, config_fn + ext)
      print("Checking '{}'...".format(test_fn), flush=True)     
      if os.path.isfile(test_fn):
        fn = maybe_replace_txt(test_fn)
        break
      #endif file exists
    #endfor extensions
    if fn is not None:
      break
    #endif found
  #endfor locations
  
  if fn is not None:
    print("Found '{}' as base startup config file.".format(fn), flush=True)
  else:
    fn = "{}.json".format(config_fn)
    print("No startup config file found in base folder", flush=True)
    os.makedirs(ct.LOCAL_CACHE, exist_ok=True)
    fn = os.path.join(ct.LOCAL_CACHE, fn)
    is_config_in_local_cache = os.path.isfile(fn)
    print("Using {}: {}".format(fn, is_config_in_local_cache), flush=True)
    config_string= os.environ.get('EE_CONFIG', './.{}.json'.format(config_fn)) # default to local .config_startup.json
    is_config_string_a_file = os.path.isfile(config_string)
    
    if is_config_in_local_cache:
      print("Found '{}' config file in local cache.".format(fn), flush=True)
    #endif local cache
    elif is_config_string_a_file:
      shutil.copy(config_string, fn)
      print("Using '{}' -> '{} as base startup config file.".format(config_string, fn), flush=True)
    #endif config string is a file
    elif len(config_string) > 3:
      # assume input is json config and we will overwrite the local cache even if it exists
      print("Attempting to process JSON '{}'...".format(config_string), flush=True)      
      config_data = json.loads(config_string)
      if isinstance(config_data, dict):
        with open(fn, 'w') as fh:
          json.dump(config_data, fh)
        print("Saved config JSON to {}".format(fn), flush=True)
      else:
        print("ERROR: EE_CONFIG '{}' is neither config file nor valid json data".format(config_string), flush=True)
        sys.exit(ct.CODE_CONFIG_ERROR)
    else:
      print("ERROR: EE_CONFIG '{}' cannot be used for startup configuration".format(config_string), flush=True)
      sys.exit(ct.CODE_CONFIG_ERROR)
      #endif cache or copy
    #endif JSON or file
  #endif default config exists or not
  return fn


def main():
  app_base_folder = os.getcwd()
  print("Core Edge Node v{} starting in {}".format(__VER__, app_base_folder), flush=True)
  
  CONFIG_FILE = 'config_startup'
  is_docker = str(os.environ.get('AINODE_DOCKER')).lower() in ["yes", "true"]
  if not is_docker:
    load_dotenv()
      
  config_file = get_config(config_fn=CONFIG_FILE)
  hostname = running_with_hostname(config_file)
  
  if hostname is not None:
    print("Hostname is '{}' - changing to ./_local_cache/{}/ cache structure".format(hostname, hostname), flush=True)
    base_folder = '_local_cache'
    app_folder = hostname
  else:
    base_folder = '.'
    app_folder = ct.LOCAL_CACHE

  l = Logger(
    lib_name='EE',
    base_folder=base_folder,
    app_folder=app_folder,
    config_file=config_file,
    max_lines=1000, 
    TF_KERAS=False
  )
  
  if l.no_folders_no_save:
    l.P("ERROR: local cache not properly configured. Note: This version is not able to use read-only systems...", color='r', boxed=True)
    sys.exit(ct.CODE_CONFIG_ERROR)
  #endif no folders
  
  if l.config_data is None or len(l.config_data) == 0:
    l.P("ERROR: config_startup.txt is not a valid json file", color='r', boxed=True)
    sys.exit(ct.CODE_CONFIG_ERROR)
  else:
    l.P("Running with config:\n{}".format(json.dumps(l.config_data, indent=4)), color='n')

  packs = l.get_packages(as_text=True, indent=4, mandatory=MANDATORY_PACKAGES)
  l.P("Current build installed packages:\n{}".format(packs))

  # DEBUG log environment
  l.P("Environment:")
  for k in os.environ:
    if k.startswith('EE_') or k.startswith('AIXP'):
      l.P("  {}={}".format(k, os.environ[k]))
  # DEBUG end log environment

  if is_docker:
    # post config setup
    docker_env = os.environ.get('AINODE_ENV')
    l.P("Docker base layer environment {}".format(docker_env))
    # test GPU overwrite
  #endif docker post config  
  
  config_box_id = get_id(log=l)
  
  try:    
    lock = l.lock_process(config_box_id)

    l.P("Starting Execution Engine Main Loop...\n\n\n.", color=ct.COLORS.MAIN)
    eng = Orchestrator(log=l)
        
    if lock is None:
      msg = "Shutdown forced due to already started local processing node with id '{}'!".format(eng.cfg_eeid)
      eng.P(msg, color='error')
      return_code = eng.forced_shutdown()
    else:
      return_code = eng.main_loop()
      
    l.p('Execution Engine v{} main loop exits with return_code: {}'.format(
      eng.__version__, return_code), color=ct.COLORS.MAIN
    )
    l.maybe_unlock_windows_file_lock(lock)
    exit_code = return_code
      
  except Exception as e:
    l.p('Execution Engine encountered an error: {}'.format(str(e)), color=ct.COLORS.MAIN)
    l.p(traceback.format_exc())
    l.maybe_unlock_windows_file_lock(lock)    
    l.P("Execution Engine exiting with post-exception code: {}".format(ct.CODE_EXCEPTION), color='r')
    exit_code = ct.CODE_EXCEPTION
  #endtry main loop startup
  return exit_code, eng
  



from naeural_core.business.base.web_app.base_web_app_plugin import BaseWebAppPlugin as BasePlugin

__VER__ = '0.0.0.0'

_CONFIG = {
  **BasePlugin.CONFIG,
  'PROCESS_DELAY': 5,

  'NGROK_ENABLED': True,
  'NGROK_DOMAIN': None,
  'NGROK_EDGE_LABEL': None,

  'PORT': None,
  'ASSETS': None,
  'SETUP_COMMANDS': [],
  'START_COMMANDS': [],
  'AUTO_START': True,

  'RUN_COMMAND': "",

  'VALIDATION_RULES': {
    **BasePlugin.CONFIG['VALIDATION_RULES']
  },
}


class NodeJsWebAppPlugin(BasePlugin):
  """
  A plugin which handles a NodeJS web app.

  Assets must be a path to a directory containing the NodeJS app.
  """

  CONFIG = _CONFIG

  def get_start_commands(self):
    super_start_commands = super(NodeJsWebAppPlugin, self).get_start_commands()
    if len(self.cfg_run_command) > 0:
      return super_start_commands + [self.cfg_run_command]
    else:
      return super_start_commands




# TODO: move all string to ratio1
class NMonConst:
  NMON_CMD_HISTORY = 'history'
  NMON_CMD_LAST_CONFIG = 'last_config'
  NMON_CMD_E2 = 'e2'
  NMON_CMD_REQUEST = 'request'
  NMON_RES_CURRENT_SERVER = 'current_server'
  NMON_RES_E2_TARGET_ID = 'e2_target_id'
  NMON_RES_E2_TARGET_ADDR = 'e2_target_addr'
  NMON_RES_NODE_HISTORY = 'node_history'
  NMON_RES_PIPELINE_INFO = 'e2_pipelines'

class _NetworkMonitorMixin:
  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)
    self.__history = self.deque(maxlen=180)
    self.__active_nodes = set()
    self.__lost_nodes = self.defaultdict(int)
    return
  
  
  @property
  def active_nodes(self):
    return self.__active_nodes
  
  @property
  def all_nodes(self):
    return self.netmon.all_nodes

  def _convert_dct_from_addr_to_eeid_index(self, dct_addr_nodes):
    dct_eeid_lst_addrs = {}
    
    for addr in dct_addr_nodes:
      eeid = self.netmon.network_node_eeid(addr=addr)
      if eeid not in dct_eeid_lst_addrs:
        dct_eeid_lst_addrs[eeid] = []
      dct_eeid_lst_addrs[eeid].append(dct_addr_nodes[addr])
    
    dct_eeid_nodes = {}
    for eeid in dct_eeid_lst_addrs:
      latest_info = sorted(dct_eeid_lst_addrs[eeid], key=lambda x:x['last_seen_sec'])[0]
      dct_eeid_nodes[eeid] = latest_info

    return dct_eeid_nodes
  
  def _add_to_history(self):
    """
    TODO: WARNING: The dict should NOT be indexed via the eeid but rather using the address
      a bad actor could use a existing eeid to overwrite the data of another node
        
    """    
    nodes = self.netmon.network_nodes_status()
    new_nodes = []
    new_nodes_addr = []
    self.__history.append(nodes) 


    if self.address_as_index:
      processed_nodes = nodes
    else:
      processed_nodes = self._convert_dct_from_addr_to_eeid_index(nodes)
    
    current_nodes = list(processed_nodes.keys())

    for node in current_nodes:
      # node is either the eeid or the addr depending on the index setting
      addr = processed_nodes[node]['address']
      if addr not in self.__active_nodes:
        if processed_nodes[node]['last_seen_sec'] < self.cfg_supervisor_alert_time:
          eeid = self.netmon.network_node_eeid(addr=addr)
          if self.cfg_log_info:
            self.P("New node {} ({}):\n{}".format(addr, eeid, self.json.dumps(nodes[addr], indent=4)))
          new_nodes.append(eeid)
          new_nodes_addr.append(addr)
        #endif
      #endif
    #endfor
    self.__active_nodes.update(new_nodes_addr)

    return processed_nodes, new_nodes
  
  
  def _supervisor_check(self):
    is_alert, nodes = False, {}
    for addr in list(self.__active_nodes):
      last_seen_ago = self.netmon.network_node_last_seen(addr=addr, as_sec=True)
      if last_seen_ago > self.cfg_supervisor_alert_time:
        if self.cfg_log_info:
          hb = self.netmon.network_node_last_heartbeat(addr=addr)              
          hb_t = hb.get(self.const.HB.CURRENT_TIME)
          ts = hb[self.const.PAYLOAD_DATA.EE_TIMESTAMP]
          tz = hb.get(self.const.PAYLOAD_DATA.EE_TIMEZONE)
          dt_local = self.log.utc_to_local(ts, tz, fmt=self.const.HB.TIMESTAMP_FORMAT)
          dt_now = self.datetime.now()
          
          elapsed = dt_now.timestamp() - dt_local.timestamp()
          
          eeid = self.netmon.network_node_eeid(addr=addr)
          self.P("Found issue with {} ({}):\n\nLast seen: {}\nStatus: {}\nHB: {}\n\n".format(
            addr, eeid, last_seen_ago, self.json.dumps(self.__history[-1][addr], indent=4),
            self.json.dumps(dict(hb_t=hb_t, ts=ts, tz=tz, elapsed=elapsed),indent=4)
          ))
        #endif debug
        uptime_sec = self.netmon.network_node_uptime(addr=addr, as_str=True)
        is_alert = True
        eeid = self.netmon.network_node_eeid(addr=addr)
        nodes[eeid] = {
            'last_seen_sec' : round(last_seen_ago, 1),
            'uptime_sec' : uptime_sec,
        }
        self.__lost_nodes[addr] += 1
      else:
        if addr in self.__lost_nodes:
          del self.__lost_nodes[addr]
      #endif is active or not
    #endfor check if any "active" is not active anymore

    for lost_addr in self.__lost_nodes:
      if self.__lost_nodes[lost_addr] > 10 and lost_addr in self.__active_nodes:
        self.__active_nodes.remove(lost_addr)
        self.P("Removed '{}' from active nodes after {} fails. Ongoing issues: {}".format(
          lost_addr, self.__lost_nodes[lost_addr], {k:v for k,v in self.__lost_nodes.items()},
        ))
      #endif
    #endfor clean lost nodes after a while
    return is_alert, nodes    
    
  
  def _get_rankings(self):
    nodes = self.__history[-1]
    if self.cfg_exclude_self:
      nodes = {k:v for k,v in nodes.items() if k != self.e2_addr}
    # TODO: change to addresses later
    eeid_nodes = self._convert_dct_from_addr_to_eeid_index(nodes)
    ranking = [(eeid, eeid_nodes[eeid]['SCORE'], eeid_nodes[eeid]['last_seen_sec']) for eeid in eeid_nodes]
    ranking = sorted(ranking, key=lambda x:x[1], reverse=True)
    return ranking  
  
  
  def _exec_netmon_request(self, target_addr, request_type, request_options={}, data={}):
    if not isinstance(request_options, dict):
      request_options = {}
    payload_params = {}
    elapsed_t = 0
    if target_addr not in self.all_nodes:
      self.P("Received `{}` request for missing node '{}'".format(
        request_type, target_addr), color='r'
      )
      return
    
    eeid = self.netmon.network_node_eeid(addr=target_addr)
    
    if request_type == NMonConst.NMON_CMD_HISTORY:
      step = request_options.get('step', 20)
      time_window_hours = request_options.get('time_window_hours', 24)
      if not isinstance(time_window_hours, int):
        time_window_hours = 24
      if not isinstance(step, int):
        step = 4
      minutes = time_window_hours * 60
      self.P("Received edge node history request for {}, step={}, hours={}".format(
        target_addr, step, minutes // 60
      ))
      start_t = self.time()
      info = self.netmon.network_node_history(
        addr=target_addr, hb_step=step, minutes=minutes,
        reverse_order=False
      )
      elapsed_t = round(self.time() - start_t, 5)
      payload_params[NMonConst.NMON_RES_NODE_HISTORY] = info
      
    elif request_type == NMonConst.NMON_CMD_LAST_CONFIG:
      self.P("Received edge node status request for '{}'".format(target_addr))
      info = self.netmon.network_node_pipelines(addr=target_addr)    
      payload_params[NMonConst.NMON_RES_PIPELINE_INFO] = info
      
    else:
      self.P("Network monitor on `{}` received invalid request type `{}` for target:address <{}:{}>".format(
        self.e2_addr, request_type, target_addr, eeid
        ), color='r'
      )
      return
    # construct the payload
    payload_params[NMonConst.NMON_CMD_REQUEST] = request_type
    payload_params[NMonConst.NMON_RES_CURRENT_SERVER] = self.e2_addr
    payload_params[NMonConst.NMON_RES_E2_TARGET_ADDR] = target_addr 
    payload_params[NMonConst.NMON_RES_E2_TARGET_ID] = eeid
    self.P("  Network monitor sending <{}> response to <{}>".format(request_type, target_addr))
    self.add_payload_by_fields(
      call_history_time=elapsed_t,
      command_params=data,
      **payload_params,
    )
    return



import os
import numpy as np

import tensorflow as tf
import tensorflow_addons as tfa

from functools import partial
from ratio1 import BaseDecentrAIObject

__VER__ = '0.2.0.0'

class TFImageDatasetsStagePreprocesserWrapper(BaseDecentrAIObject):

  def __init__(self, log, batch_size,
               image_height, image_width,
               train_dataset_directory,
               dev_dataset_directory=None,
               test_dataset_directory=None,
               files_extension=None,
               apply_preprocess=True,
               change_stage_points=None,
               resize_with_pad=False,
               **kwargs):

    """
    Parameters:
    -----------
    batch_size : int, mandatory
      Batch size

    image_height : int, mandatory
      Input image height (the images may have any height as the preprocesser will resize them to this value)

    image_width : int, mandatory
      Input image width (the images may have any width as the preprocesser will resize them to this value)

    train_dataset_directory : str, mandatory
      Directory where the training images are stored per each class
      Example: (`train_dataset_directory` = '.../.../.../TRAIN')
        .../.../.../TRAIN
          class1
            img11.png
            img12.png
            ...
          class2
            img21.png
            img22.png

    dev_dataset_directory : str, optional
      Directory where the dev images are stored per each class (same structure as the training images)
      The default is None - no dev dataset

    test_dataset_directory : str, optional
      Directory where the test images are stored per each class (same structure as the training images)
      The default is None - no test dataset

    files_extension : str, optional
      The extension of image files (.png, .jpg, .jpeg)
      The default is None - .png

    apply_preprocess : bool, optional
      Whether to activate the staging preprocessing
      The default is True

    change_stage_points : dict, optional
      Dictionary that specifies the pairing between the epoch and the used preprocessing stage
      Example: (`change_stage_points` = {3:1, 10:2, 20: 1, 30: 2}
        The preprocesser will not apply preprocessing functions until epoch 3; then between
        epochs 3 and 9 will apply stage 1 preprocessing functions; then between epochs 10 and 10
        will apply stage 2 preprocessing functions and so on.
      The default is None - {3:1, 10:2}

    resize_with_pad : bool, optional
      Whether the resize to (`image_height`, `image_width`) keeps the aspect ratio or not.
      The default is False

    Public methods:
    ---------------
    - get_train_tf_dataset()
    - get_dev_tf_dataset()
    - get_test_tf_dataset()
    - get_dct_classes()
    - train_epoch_changed(epoch=epoch)
    - get_original_dev_dataset()
    - get_original_test_dataset()
    - get_stage1_dev_dataset(times=x)
    - get_stage2_dev_dataset(times=x)
    """

    if files_extension is None:
      files_extension = '.png'

    if change_stage_points is None:
      change_stage_points = {3:1, 10:2}

    self.version = __VER__
    super().__init__(log=log, prefix_log='[IDPW]', **kwargs)

    self._train_preproc = TFImageDatasetStagePreprocesser(
      log=log, batch_size=batch_size,
      image_height=image_height, image_width=image_width,
      dataset_directory=train_dataset_directory,
      files_extension=files_extension,
      apply_preprocess=apply_preprocess,
      change_stage_points=change_stage_points,
      resize_with_pad=resize_with_pad,
      **kwargs
    )

    self._dev_preproc = None
    if dev_dataset_directory is not None:
      self._dev_preproc = TFImageDatasetStagePreprocesser(
        log=log, batch_size=batch_size,
        image_height=image_height, image_width=image_width,
        dataset_directory=dev_dataset_directory,
        files_extension=files_extension,
        apply_preprocess=apply_preprocess,
        change_stage_points={1:1, 2:2},
        resize_with_pad=resize_with_pad,
        **kwargs
      )
    #endif

    self._test_preproc = None
    if test_dataset_directory is not None:
      self._test_preproc = TFImageDatasetStagePreprocesser(
        log=log, batch_size=batch_size,
        image_height=image_height, image_width=image_width,
        dataset_directory=test_dataset_directory,
        files_extension=files_extension,
        apply_preprocess=False,
        change_stage_points={},
        resize_with_pad=resize_with_pad,
        **kwargs
      )
    #endif

    return

  def get_train_tf_dataset(self):
    return self._train_preproc()

  def get_dev_tf_dataset(self):
    if self._dev_preproc:
      return self._dev_preproc()

  def get_test_tf_dataset(self):
    if self._test_preproc:
      return self._test_preproc()

  def get_dct_classes(self):
    return self._train_preproc.dct_classes

  def get_train_preprocesser(self):
    return self._train_preproc

  def get_dev_preprocesser(self):
    return self._dev_preproc

  def get_test_preprocesser(self):
    return self._test_preproc

  def train_epoch_changed(self, epoch):
    self._train_preproc.epoch_changed(epoch)

  def _append_to_results(self, results, batch):
    if results is None:
      results = batch
    else:
      results = [tf.concat([results[i], batch[i]], 0) for i in range(len(results))]
    return results

  def _full_dataset(self, ds='dev'):
    assert ds.lower() in ['dev', 'test']
    if ds == 'dev':
      tf_ds = self.get_dev_tf_dataset()
    else:
      tf_ds = self.get_test_tf_dataset()

    results = None
    if tf_ds is not None:
      for batch in tf_ds:
        results = self._append_to_results(results, batch)

    return results

  def _stage_dataset(self, stage, ds='dev', times=1):
    results = None
    if self._dev_preproc:
      self._dev_preproc.epoch_changed(stage)
      for _ in range(times):
        crt_results = self._full_dataset(ds)
        results = self._append_to_results(results, crt_results)

    return results

  def get_original_dev_dataset(self):
    return self._full_dataset('dev')

  def get_stage1_dev_dataset(self, times=1):
    """
    Augments the dev dataset using stage 1 preprocessing functions x times
    """
    return self._stage_dataset(stage=1, ds='dev', times=times)

  def get_stage2_dev_dataset(self, times=1):
    """
    Augments the dev dataset using stage 2 preprocessing functions x times
    """
    return self._stage_dataset(stage=2, ds='dev', times=times)

  def get_original_test_dataset(self):
    return self._full_dataset('test')


class TFImageDatasetStagePreprocesser(BaseDecentrAIObject):

  def __init__(self, log, batch_size,
               image_height, image_width,
               dataset_directory,
               files_extension=None,
               apply_preprocess=True,
               change_stage_points={3:1, 10:2},
               resize_with_pad=False,
               **kwargs):
    if files_extension is None:
      files_extension = '.png'
    assert files_extension in ['.png', '.jpg', '.jpeg']
    self.version = __VER__
    super().__init__(log=log, prefix_log='[IDP]', **kwargs)
    self.batch_size = batch_size
    self.image_height = image_height
    self.image_width = image_width
    self.change_stage_points = change_stage_points
    self.dataset_directory = dataset_directory
    self.apply_preprocess = apply_preprocess
    self.resize_with_pad = resize_with_pad
    self.files_extension = files_extension

    self.current_stage = 0
    self.tf_current_stage = tf.Variable(initial_value=0)
    self.epoch = None
    self.epoch_changed(0)
    self.dct_classes = None

    ########### TODO does not work yet with `get_stage_image`
    ### each stage is a list of preprocessors
    # self.stages = {s : [] for s in range(self.nr_stages)}
    #
    # self.stages[1] = [
    #   self.preprocess_flip_left_right,
    #   partial(self.preprocess_brightness, max_delta=0.5),
    # ]
    #
    # self.stages[2] = self.stages[1] + [
    #   partial(self.preprocess_crop_v2, min_crop_prc=0.05, max_crop_prc=0.3),
    #   partial(self.preprocess_rotation, max_degrees=20)
    # ]
    ############

    self.func1_1 = self.preprocess_flip_left_right
    self.func1_2 = partial(self.preprocess_brightness, max_delta=0.5)
    self.func2_1 = partial(self.preprocess_crop_v2, min_crop_prc=0.05, max_crop_prc=0.25)
    self.func2_2 = partial(self.preprocess_rotation, max_degrees=15)


    self.P("Preprocessing images from '{}' directory ...".format(self.dataset_directory))

    for _, l, _ in os.walk(self.dataset_directory):
      self.labels = l
      break

    self.P("  Found {} classes: {}".format(len(self.labels), self.labels))
    self.dct_classes = {i : self.labels[i] for i in range(len(self.labels))}

    return

  # def get_stage_image(self, image):
  #   lst_methods = self.stages[self.current_stage]
  #   loop_vars = (tf.constant(0), image)
  #   cond = lambda i, tf_img: i < len(lst_methods)
  #   body = lambda i, tf_img: (i+1, tf.cond(tf.random.uniform([], 0, 100) >= 50, lambda: tf_img, lambda: (tf.gather(lst_methods, i))(tf_img)))
  #   final = tf.while_loop(cond, body, loop_vars)
  #   image = final[1]
  #   return image

  def reset_stage(self):
    self.tf_current_stage.assign(0)
    self.current_stage = 0
    return

  def set_stage(self, stage):
    self.tf_current_stage.assign(stage)
    self.current_stage = stage
    return

  def set_max_stage(self):
    max_stage = max(self.change_stage_points.values())
    self.set_stage(max_stage)
    return

  def _get_stage_one_image(self, image):
    # self.log.start_timer('stage_one_image')
    image = tf.cond(tf.random.uniform([], 0, 100) >= 50, lambda: image, lambda: self.func1_1(image))
    image = tf.cond(tf.random.uniform([], 0, 100) >= 50, lambda: image, lambda: self.func1_2(image))
    # self.log.end_timer('stage_one_image')
    return image

  def _get_stage_two_image(self, image):
    # self.log.start_timer('stage_two_image')
    image = self._get_stage_one_image(image)
    image = tf.cond(tf.random.uniform([], 0, 100) >= 50, lambda: image, lambda: self.func2_1(image))
    image = tf.cond(tf.random.uniform([], 0, 100) >= 50, lambda: image, lambda: self.func2_2(image))
    # self.log.end_timer('stage_two_image')
    return image

  @tf.function
  def _staged_preprocess_image(self, image, apply_preprocess):
    # tf.print("Current stage", self.tf_current_stage)
    def true_fn(tf_img):
      tf_img = tf.cond(tf.math.equal(self.tf_current_stage, 1), lambda: self._get_stage_one_image(tf_img), lambda: tf_img)
      tf_img = tf.cond(tf.math.equal(self.tf_current_stage, 2), lambda: self._get_stage_two_image(tf_img), lambda: tf_img)
      return tf_img

    image = tf.cond(
      tf.constant(apply_preprocess),
      lambda: true_fn(image),
      lambda: image
    )

    return image

  def epoch_changed(self, epoch):
    self.epoch = epoch
    stage = self.change_stage_points.get(self.epoch, None)
    if stage:
      self.set_stage(stage)
    return

  def __call__(self):
    return self.tf_dataset()

  def preprocess_flip_left_right(self, image):
    # tf.print("DEBUG: preprocess_flip_left_right")
    return tf.image.flip_left_right(image)

  def preprocess_crop(self, image, top_prc=0.0, left_prc=0.0, bottom_prc=0.0, right_prc=0.0):
    # tf.print("DEBUG: preprocess_crop")
    tf_img_shape = tf.shape(image)
    tf_img_hw = tf.cast(tf.slice(tf_img_shape, [0], [2]), 'float32')
    kwargs = self._tf_get_crop_to_bounding_box_params(
      tf_img_hw,
      top_prc=top_prc, left_prc=left_prc,
      bottom_prc=bottom_prc, right_prc=right_prc
    )

    return tf.image.crop_to_bounding_box(image=image, **kwargs)

  def preprocess_brightness(self, image, max_delta=0.5):
    # tf.print("DEBUG: preprocess_brightness")
    return tf.image.random_brightness(image=image, max_delta=max_delta)

  def preprocess_rotation(self, image, max_degrees=15):
    # tf.print("DEBUG: preprocess_rotation")
    if tf.math.greater(tf.abs(max_degrees), 20):
      tf.print("WARNING! In rotation preproc, |max_degrees| is greater than 20. Please be sure that you want higher rotation")

    degrees = tf.random.uniform([], -max_degrees, max_degrees, seed=None)

    ### Info: we use tensorflow-addons!
    return tfa.image.rotate(
      images=image,
      angles=self._degrees_to_radians(degrees),
    )

  def preprocess_crop_v2(self, image, min_crop_prc=0.05, max_crop_prc=0.3):
    if tf.math.greater(max_crop_prc, 0.5):
      tf.print("WARNING! In crop preproc, max_proc_prc is greater than 0.5. Please be sure that you want to crop more than half of the images")

    prc = tf.random.uniform([], min_crop_prc, max_crop_prc)
    top_prc, left_prc, bottom_prc, right_prc = self._get_cropping_percents(prc)

    return self.preprocess_crop(
      image=image,
      top_prc=top_prc,
      left_prc=left_prc,
      bottom_prc=bottom_prc,
      right_prc=right_prc
    )

  @staticmethod
  def _get_cropping_percents(prc):
    if False:
      # 0 - crop top
      # 1 - crop left
      # 2 - crop bottom
      # 3 - crop right
      # 4 - crop top-left
      # 5 - crop bottom-right

      tf_option = tf.random.uniform([], 0, 6, dtype=tf.dtypes.int32)

      if tf.equal(tf_option, 0):
        t, l, b, r = prc, 0.0, 0.0, 0.0
      elif tf.equal(tf_option, 1):
        t, l, b, r = 0.0, prc, 0.0, 0.0
      elif tf.equal(tf_option, 2):
        t, l, b, r = 0.0, 0.0, prc, 0.0
      elif tf.equal(tf_option, 3):
        t, l, b, r = 0.0, 0.0, 0.0, prc
      elif tf.equal(tf_option, 4):
        t, l, b, r = prc, prc, 0.0, 0.0
      elif tf.equal(tf_option, 5):
        t, l, b, r = 0.0, 0.0, prc, prc
      else:
        t, l, b, r = 0.0, 0.0, 0.0, 0.0
    #endif

    t, l, b, r = 0.0, 0.0, prc, 0.0
    return t,l,b,r


  @staticmethod
  def _get_crop_to_bounding_box_params(H, W, top_prc=0, left_prc=0, bottom_prc=0, right_prc=0):
    top = int(top_prc * H)
    left = int(left_prc * W)
    bottom = H - int(bottom_prc * H)
    right = W - int(right_prc * W)

    target_H = bottom - top
    target_W = right - left

    return {
      'offset_height' : top,
      'offset_width' : left,
      'target_height' : target_H,
      'target_width' : target_W
    }

  @staticmethod
  def _tf_get_crop_to_bounding_box_params(tf_img_hw, top_prc, left_prc, bottom_prc, right_prc):
    tf_procents_tl = tf.stack([top_prc, left_prc])
    tf_procents_br = tf.stack([bottom_prc, right_prc])

    tf_tl = tf.cast(tf.round(tf.math.multiply(tf_procents_tl, tf_img_hw)), 'int32')
    tf_br = tf.cast(tf_img_hw - tf.round(tf.math.multiply(tf_procents_br, tf_img_hw)), 'int32')

    target_hw = tf_br - tf_tl

    return {
      'offset_height' : tf_tl[0],
      'offset_width' : tf_tl[1],
      'target_height' : target_hw[0],
      'target_width' : target_hw[1]
    }

  @staticmethod
  def _degrees_to_radians(degrees):
    return degrees * np.pi / 180


  @tf.function
  def _resize(self, image):
    image = tf.cond(
      tf.constant(self.resize_with_pad),
      lambda: tf.image.resize_with_pad(
        image,
        target_height=self.image_height,
        target_width=self.image_width,
      ),
      lambda: tf.image.resize(
        image,
        size=[self.image_height, self.image_width]
      )
    )
    return image

  @tf.function
  def read_staged_preprocess_image(self, file_name):
    # tf.print("!!!READ_IMAGE!!!")
    # self.log.start_timer('read_image')
    image = tf.io.read_file(file_name)
    image = tf.image.decode_jpeg(image, channels=3)
    # self.log.end_timer('read_image')
    # image = tf.cond(tf.constant(apply_preprocess), lambda: self.get_stage_image(image), lambda: image)
    # self.log.start_timer('maybe_process_image')
    image = self._staged_preprocess_image(image=image, apply_preprocess=self.apply_preprocess)
    # self.log.end_timer('maybe_process_image')
    # self.log.start_timer('resize')
    image = self._resize(image)
    # self.log.end_timer('resize')
    # self.log.start_timer('cast_and_label')
    image = tf.cast(image, tf.uint8)
    label = tf.strings.split(file_name, os.sep)[-2]
    label = tf.squeeze(tf.where(tf.equal(label, self.labels)), axis=1)
    # self.log.end_timer('cast_and_label')
    return image, label#, file_name
  # enddef

  def tf_dataset(self):
    p = '{}/*/*{}'.format(self.dataset_directory, self.files_extension)
    list_ds = tf.data.Dataset.list_files(p)
    labeled_ds = list_ds.map(self.read_staged_preprocess_image)
    labeled_ds = labeled_ds.batch(self.batch_size)
    return labeled_ds

if __name__ == '__main__':
  import matplotlib.pyplot as plt
  from naeural_core import Logger

  log = Logger(
    lib_name='GIDS',
    base_folder='.',
    app_folder='_local_cache',
    TF_KERAS=False
  )

  url = 'https://www.dropbox.com/sh/6iiec6xryd1nrxz/AADO2xCx_6gMcnp51TlUfPX7a?dl=1'
  ds_name = 'hard_cover_datasets'

  saved_files, _ =log.maybe_download(
    url=url,
    fn=ds_name,
    target='data',
    unzip=True
  )

  BATCH_SIZE = 9
  IMG_HEIGHT, IMG_WIDTH = 224, 112
  data_dir = saved_files[0]

  train_data_dir = os.path.join(data_dir, 'TRAIN')
  dev_data_dir = os.path.join(data_dir, 'DEV')
  test_data_dir = os.path.join(data_dir, 'TEST')

  preproc_w = TFImageDatasetsStagePreprocesserWrapper(
    log=log,
    batch_size=BATCH_SIZE, image_height=IMG_HEIGHT, image_width=IMG_WIDTH,
    train_dataset_directory=train_data_dir,
    dev_dataset_directory=dev_data_dir,
    test_dataset_directory=test_data_dir,
    apply_preprocess=True,
    change_stage_points={3: 1, 10: 2},
    resize_with_pad=False
  )

  dev_results = preproc_w.get_original_dev_dataset()
  dev_results_2 = preproc_w.get_stage1_dev_dataset(times=4)

  if False:
    for epoch in [2, 3, 4, 10, 11]:
      print("Epoch {}".format(epoch))
      preproc_w.train_epoch_changed(epoch)
      ds_trn = preproc_w.get_train_tf_dataset()
      images,labels,fns = next(iter(ds_trn))
      str_epoch = 'E{}'.format(epoch)
      if True:
        plt.figure(figsize=(10, 10))
        for i in range(9):
          label = preproc_w.get_dct_classes()[labels[i,0].numpy()]
          fn = fns[i].numpy().decode().split('/')[-1]
          ax = plt.subplot(3, 3, i + 1)
          plt.imshow(images[i].numpy().astype("uint8"))
          plt.title(label + '-' + fn + '-' + str_epoch)
          plt.axis("off")
        plt.show()




import numpy as np

import plotly.graph_objects as go
import dash
import dash_core_components as dcc
import dash_html_components as html
from dash.dependencies import Input, Output

class AbstractAttentionVisualizer(object):
    '''
    Parameters:
        input_sequence: tensor with shape (batch_size, input_size, 1)
        output_sequence: tensor with shape (batch_size, output_size, 1)
        attention_matrix: tensor with shape (batch_size, output_size, input_size)
    '''
    def __init__(self, input_sequence, output_sequence, attention_matrix):
        self.input_sequence = input_sequence
        self.output_sequence = output_sequence
        self.attention_matrix = attention_matrix
        self.data_id = 0
        self.output_item_id = 0
        self._input_check()

    def _input_check(self):
        if self.input_sequence.shape[1] != self.attention_matrix.shape[2] or \
            self.output_sequence.shape[1] != self.attention_matrix.shape[1]:
            raise ValueError('Tensor args shape does not match!')

    def select_data(self, id):
        if id < 0 or id >= self.input_sequence.shape[0]:
            return False

        self.data_id = id
        return True

    def select_output_item(self, id):
        if id < 0 or id >= self.output_sequence.shape[1]:
            return False

        self.output_item_id = id
        return True

    def get_input_sequence(self):
        return self.input_sequence[self.data_id, :, 0]

    def get_output_sequence(self):
        return self.output_sequence[self.data_id, :, 0]

    def get_input_ids(self):
        return list(range(self.get_input_sequence().shape[0]))

    def get_output_ids(self):
        return list(range(self.get_output_sequence().shape[0]))

    def get_attention(self):
        return self.attention_matrix[self.data_id, self.output_item_id:self.output_item_id+1]

    def get_full_attention(self):
        return self.attention_matrix[self.data_id]

    def get_most_important_input(self, window=3):
        sequence = self.get_attention()[0]
        minval = min(sequence)
        maxval = max(sequence)
        threshold = minval + 0.75 * (maxval - minval)

        out_size = sequence.shape[-1]
        max_quants = 0
        max_total = 0
        id = 0
        for i in range(0, out_size-window):
            quants = sum(sequence[i:i+window] > threshold)
            total = sum(sequence[i:i+window])
            if quants > max_quants:
                max_quants, max_total = quants, total
                id = i + window // 2
            elif quants == max_quants and total > max_total:
                max_total = total
                id = i + window // 2

        return id

class DashLensAttentionVisualizer(AbstractAttentionVisualizer):
    '''
    Dash(Plotly) implementation for the Lens Attention Visualizer
    Parameters:
        input_sequence: tensor with shape (batch_size, input_size, 1)
        output_sequence: tensor with shape (batch_size, output_size, 1)
        attention_matrix: tensor with shape (batch_size, output_size, input_size)

        input_name (opt): historical figure component identifier
        output_name (opt): forecasted figure component identifier
        attention_name (opt): attention figure component identifier

        important_attn_text (opt): annotation text for most influential window in attention
        selected_output_text (opt): annotation text for selected forecasted day

        input_title (opt): figure title for historical plot
        input_xlabel (opt): x-axis label for historical plot
        input_ylabel (opt): y-axis label for historical plot

        output_title (opt): figure title for forecasted plot
        output_xlabel (opt): x-axis label for forecasted plot
        output_ylabel (opt): y-axis label for forecasted plot

        attention_title (opt): figure title for attention plot
        attention_xlabel (opt): x-axis label for attention plot
        attention_ylabel (opt): y-axis label for attention plot

        colorscale (opt): colorscale option for the attention colorbar
    '''
    def __init__(self, input_sequence, output_sequence, attention_matrix,
                input_name='Historical', output_name='Forecasted', attention_name='Attention',
                important_attn_text='Most important', selected_output_text='Selected day',
                input_title='Historical data', input_xlabel='Historical day', input_ylabel='Sold quantity',
                output_title='Forecasted data', output_xlabel='Forecasted day', output_ylabel='Expected sell quantity',
                attention_title='Historical influence', attention_xlabel='Historical day', attention_ylabel='Forecasted day',
                colorscale='Viridis'):
        super().__init__(input_sequence, output_sequence, attention_matrix)
        self.input_name = input_name
        self.output_name = output_name
        self.attention_name = attention_name

        self.important_attn_text = important_attn_text
        self.selected_output_text = selected_output_text

        self.input_title = input_title
        self.input_xlabel, self.input_ylabel = input_xlabel, input_ylabel

        self.output_title = output_title
        self.output_xlabel, self.output_ylabel = output_xlabel, output_ylabel

        self.attention_title = attention_title
        self.attention_xlabel, self.attention_ylabel = attention_xlabel, attention_ylabel

        self.colorscale = colorscale

        self.dropdown_id = "dropdown"

        self.app = self._build_app()

    def _build_historical(self):
        return go.Figure(
            data=[go.Scatter(
                x=self.get_input_ids(),
                y=self.get_input_sequence(),
                name=self.input_name
            )],
            layout=go.Layout(
                title=dict(
                    text=self.input_title
                ),
                xaxis=dict(
                    title=self.input_xlabel
                ),
                yaxis=dict(
                    title=self.input_ylabel
                )
            )
        )

    def _build_forecasted(self):
        return go.Figure(
            data=[go.Scatter(
                x=self.get_output_ids(),
                y=self.get_output_sequence(),
                name=self.output_name
            )],
            layout=go.Layout(
                title=dict(
                    text=self.output_title
                ),
                xaxis=dict(
                    title=self.output_xlabel
                ),
                yaxis=dict(
                    title=self.output_ylabel
                ),
                annotations=[
                    dict(
                        x=self.output_item_id,
                        y=self.get_output_sequence()[self.output_item_id],
                        xref='x',
                        yref='y',
                        ax=0,
                        ay=30,
                        text=self.selected_output_text,
                        font=dict(
                            color="black",
                            size=12
                        ),
                        arrowcolor="black",
                        arrowsize=2,
                        arrowwidth=1,
                        arrowhead=1
                    )
                ]
            )
        )

    def _build_attention(self):
        return go.Figure(
            data=[go.Heatmap(
                x=self.get_input_ids(),
                y=[self.get_output_ids()[self.output_item_id]],
                z=self.get_attention(),
                colorscale=self.colorscale,
                name=self.attention_name
            )],
            layout=go.Layout(
                title=dict(
                    text=self.attention_title
                ),
                xaxis=dict(
                    title=self.attention_xlabel
                ),
                yaxis=dict(
                    title=self.attention_ylabel,
                    nticks=1
                ),
                annotations=[
                    dict(
                        x=self.get_most_important_input()/self.input_sequence.shape[1]+ \
                            1/(2*self.input_sequence.shape[1]),
                        y=0,
                        xref='paper',
                        yref='paper',
                        ax=0,
                        ay=30,
                        text=self.important_attn_text,
                        font=dict(
                            color="black",
                            size=12
                        ),
                        arrowcolor="black",
                        arrowsize=2,
                        arrowwidth=1,
                        arrowhead=1
                    )
                ]
            )
        )

    def _update_historical(self, dropdown_value):
        if dropdown_value is not None:
            self.select_data(dropdown_value)

        return self._build_historical()

    def _update_forecasted(self, dropdown_value, forecast_clickData):
        if dropdown_value is not None:
            self.select_data(dropdown_value)

        if forecast_clickData is not None:
            self.select_output_item(forecast_clickData['points'][0]['pointIndex'])

        return self._build_forecasted()

    def _update_attention(self, dropdown_value, forecast_clickData):
        if dropdown_value is not None:
            self.select_data(dropdown_value)

        if forecast_clickData is not None:
            self.select_output_item(forecast_clickData['points'][0]['pointIndex'])

        return self._build_attention()

    def _build_app(self):
        app = dash.Dash(__name__)
        app.layout = html.Div(
            children=[
                dcc.Dropdown(
                    id=self.dropdown_id,
                    options=[{'label':f'Series {i}', 'value': i} for i in range(self.input_sequence.shape[0])],
                    value=self.data_id
                ),
                html.Div(
                    children=[
                        dcc.Graph(
                            id=self.input_name,
                            figure=self._build_historical(),
                            style={'width':'50%','display':'inline-block'}
                        ),
                        dcc.Graph(
                            id=self.output_name,
                            figure=self._build_forecasted(),
                            style={'width':'50%','display':'inline-block'}
                        )
                ]),
                html.Div(
                    children=[
                        dcc.Graph(
                            id=self.attention_name,
                            figure=self._build_attention(),
                            style={'width':'50%','display':'inline-block'}
                        )
                ])
            ]
        )

        app.callback(
            Output(self.input_name, 'figure'),
            [Input(self.dropdown_id, 'value')]
        )(self._update_historical)
        app.callback(
            Output(self.output_name, 'figure'),
            [Input(self.dropdown_id, 'value'),
            Input(self.output_name, 'clickData')]
        )(self._update_forecasted)
        app.callback(
            Output(self.attention_name, 'figure'),
            [Input(self.dropdown_id, 'value'),
            Input(self.output_name, 'clickData')]
        )(self._update_attention)

        return app

    def show_plot(self):
        self.app.run_server()

def generate_lens_data_example(batch_size=10, historical_size=180, forecasted_size=15,
                            historical_range=(5,10), forecasted_range=(5,10), attn_range=(0,10)):
    historical = historical_range[0] + np.random.rand(batch_size, historical_size, 1) * (historical_range[1] - historical_range[0])
    forecasted = forecasted_range[0] + np.random.rand(batch_size, forecasted_size, 1) * (forecasted_range[1] - forecasted_range[0])
    attention = attn_range[0] + np.random.rand(batch_size, forecasted_size, historical_size) * (attn_range[1] - attn_range[0])

    return historical, forecasted, attention

if __name__ == "__main__":
    h, f, a = generate_lens_data_example()
    app = DashLensAttentionVisualizer(h, f, a)
    app.show_plot()



import random
import codecs

from naeural_core import Logger
from time import time


if __name__ == '__main__':
  l = Logger('THRP', base_folder='.', app_folder='_cache')
  
  TESTS = 4000

  phrases = ['test {} '.format(x) * 10 for x in range(100)]
  buffer = []
  
  def _p(s):
    buffer.append(s)
    print(s, flush=True)
    fn = '_local_cache/_logs/test.txt'
    if False:
      with codecs.open(fn, 'w', "utf-8") as f:
        for line in buffer:
          f.write("{}\n".format(line))
    else:
      with open(fn, 'w') as f:
        f.writelines(buffer)
    return

  start1=time()
  for i in range(TESTS):
    l.start_timer('_p')
    _p(random.choice(phrases))
    l.end_timer('_p')
  end1=time()

  start2=time()
  for i in range(TESTS):
    l.start_timer('.P')
    l.P(random.choice(phrases))
    l.end_timer('.P')
  end2=time()

  l.P("_p total time: {:.2f}s".format(end1-start1))
  l.P(".P total time: {:.2f}s".format(end2-start2))
  l.show_timers()
  
# N1_L05 ipython 18%
# [THRP][2022-05-28 23:10:52]  _p = 0.0014s, max: 0.0041s, last: 0.0021s, cnt: 999
# [THRP][2022-05-28 23:10:52]  .P = 0.0017s, max: 0.0110s, last: 0.0023s, cnt: 999

# N1_L05 ipython 16%
# [THRP][2022-05-29 10:13:28]  _p = 0.0037s, max: 0.0352s, last: 0.0048s, cnt: 3999
# [THRP][2022-05-29 10:13:28]  .P = 0.0044s, max: 0.0272s, last: 0.0063s, cnt: 3999

# N1_L05 cmd 10%
# [THRP][2022-05-29 10:14:48]  _p = 0.0034s, max: 0.0240s, last: 0.0067s, cnt: 3999
# [THRP][2022-05-29 10:14:48]  .P = 0.0038s, max: 0.0292s, last: 0.0069s, cnt: 3999

# N2_L80 cmd 8%
# N2_L80 ipython 10%




import platform
import torch as th
import torchvision as tv

from naeural_core.serving.base.backends.model_backend_wrapper import ModelBackendWrapper
from naeural_core.serving.base.base_serving_process import ModelServingProcess as BaseServingProcess
from naeural_core.serving.mixins_base.th_utils import _ThUtilsMixin
from naeural_core.serving.mixins_base.ths_mixin import TorchscriptMixin
from naeural_core.serving.mixins_base.trt_mixin import TensortRTMixin
from naeural_core.serving.mixins_base.onnx_mixin import ONNXMixin
from naeural_core.serving.mixins_base.openvino_mixin import OpenVINOMixin

from naeural_core.utils.plugins_base.plugin_base_utils import NestedDotDict

BACKEND_PRIORITY = NestedDotDict({
  'ARM64_CPU' : ['onnx', 'openvino', 'ths'],
  'ARM64_GPU' : ['trt', 'ths'],
  'AMD64_CPU' : ['openvino', 'onnx', 'ths'],
  'AMD64_GPU' : ['trt', 'ths'],
})

BACKEND_REQUIRES_CPU = ['onnx', 'openvino']


_CONFIG = {
  **BaseServingProcess.CONFIG,

  "IMAGE_HW"                    : None,
  "WARMUP_COUNT"                : 3,
  "USE_AMP"                     : False,
  "USE_FP16"                    : None,
  "DEFAULT_DEVICE"              : None,
  "URL"                         : None,
  "DEBUG_TIMERS"                : False,
  "MAX_BATCH_FIRST_STAGE"       : None,
  "MAX_BATCH_SECOND_STAGE"      : None,

  "MODEL_WEIGHTS_FILENAME"      : None,
  "MODEL_CLASSES_FILENAME"      : None,

  "MODEL_ONNX_FILENAME"         : None,
  "ONNX_URL"                    : None,
  "MODEL_TRT_FILENAME"          : None,
  "TRT_URL"                     : None,
  "BACKEND"                     : None,

  "SECOND_STAGE_MODEL_WEIGHTS_FILENAME" : None,
  "SECOND_STAGE_MODEL_CLASSES_FILENAME" : None,

  "CUDNN_BENCHMARK"             : False,

  'VALIDATION_RULES': {
    **BaseServingProcess.CONFIG['VALIDATION_RULES'],
  },
}


class UnifiedFirstStage(
  BaseServingProcess,
  _ThUtilsMixin,
  TorchscriptMixin,
  TensortRTMixin,
  ONNXMixin,
  OpenVINOMixin,
):

  CONFIG = _CONFIG

  def __init__(self, **kwargs):
    self.default_input_shape = None
    self.class_names = None
    self.graph_config = {}
    self._platform_backend_priority = None
    self._str_dev = None

    super(UnifiedFirstStage, self).__init__(**kwargs)
    return

  @property
  def th(self):
    return th

  @property
  def tv(self):
    return tv

  @property
  def cfg_input_size(self):
    return self.cfg_image_hw

  @property
  def cfg_fp16(self):
    # Use the USE_FP16 value if that was specified in the config.
    # If it wasn't specified and we're using the CPU don't use
    # fp16 as that will crash pytorch. If we're using CUDA
    # default to fp16.
    if self.cfg_use_fp16 is not None:
      return self.cfg_use_fp16
    if self.dev.type == 'cpu':
      return False
    return True

  @property
  def th_dtype(self):
    ### TODO: Maybe we have uint as input
    return th.float16 if self.cfg_fp16 else th.float32

  @property
  def get_model_weights_filename(self):
    # This is done in order for us to be able to have more control over the weights filename
    return self.cfg_model_weights_filename

  @property
  def get_model_classes_filename(self):
    return self.cfg_model_classes_filename

  @property
  def get_model_trt_filename(self):
    return self.cfg_model_trt_filename

  @property
  def get_trt_url(self):
    return self.cfg_trt_url

  @property
  def get_model_onnx_filename(self):
    return self.cfg_model_onnx_filename

  @property
  def get_onnx_url(self):
    return self.cfg_onnx_url

  @property
  def get_url(self):
    return self.cfg_url

  def get_device_string(self):
    """
    Get the device string for the torch device used by this serving.

    Parameters
    ----------
    None

    Returns
    str - the torch device string used
    """
    return self._str_dev

  def get_saved_model_fn(self):
    fn_saved_model = self.get_model_weights_filename if self.get_model_weights_filename is not None else self.server_name + '_weights.pt'
    return fn_saved_model

  def get_saved_classes_fn(self):
    fn_saved_classes = self.get_model_classes_filename if self.get_model_classes_filename is not None else self.server_name + '_classes.txt'
    return fn_saved_classes

  def has_device(self, dev=None):
    if dev is None:
      dev = th.device(self.get_device_string())
    elif isinstance(dev, str):
      dev = th.device(dev)
    dev_idx = 0 if dev.index is None else dev.index
    if dev.type == 'cpu' or (th.cuda.is_available() and th.cuda.device_count() > dev_idx):
      return True
    return False

  def _stop_timer(self, tmr, periodic=False):
    if self.cfg_debug_timers:
      th.cuda.synchronize()
    return super()._stop_timer(tmr, periodic=periodic)

  def _get_platform_backend_order(self):
    """
    Produces a list of supported backends, ordered by preference (from
    most preferable to least), given the current platform and torch device.

    Parameters
    ----------
    None

    Returns
    -------
    ret - Sequence[str] - a list of backend names to be used
    """
    machine = platform.machine().lower()
    # Check for x86/x86_64/amd64
    is_x86 = machine.startswith('x86_64') or machine.startswith('amd64')
    # Check for arm{32|64} or aarch{32|64}
    is_arm64 = machine.startswith('arm64') or machine.startswith('aarch64')
    is_cpu = (self.dev.type == 'cpu')
    if is_cpu:
      if is_x86:
        return BACKEND_PRIORITY.AMD64_CPU
      if is_arm64:
        return BACKEND_PRIORITY.ARM64_CPU

    if is_x86:
      return BACKEND_PRIORITY.AMD64_GPU
    if is_arm64:
      return BACKEND_PRIORITY.ARM64_GPU
    # TODO: we might have more platforms and this code is a bit too specific.
    raise ValueError('Unexpected platform')

  def _setup_device_string(self):
    """
    Initializes the device string of this serving based on the
    configuration and platform. If DEFAULT_DEVICE was explicitly
    set in the config, we use that. If a backend that needs the
    CPU was explicitly and DEFAULT_DEVICE was not specified, we
    will use the CPU device. Otherwise, if the platform has a
    CUDA device we will use cuda:0 or fall back to the CPU.

    Parameters
    ----------
    None

    Returns
    -------
    None
    """
    has_cuda = (th.cuda.is_available() and th.cuda.device_count() > 0)
    if self.cfg_default_device is not None:
      self._str_dev = self.cfg_default_device.lower()
      return
    elif has_cuda:
      self.P("Machine support CUDA, using cuda:0 as the default device", color='b')
      self._str_dev = 'cuda:0'
    else:
      self.P("Machine does not support CUDA, falling back to CPU", color='r')
      self._str_dev = 'cpu'
    #endif select default device

    if self.cfg_backend is not None:
      if self.cfg_backend in BACKEND_REQUIRES_CPU:
        self._str_dev = 'cpu'
        self.P("Forcing device to CPU for backend", c='r')
    #endif force device for backend
    return

  def _setup_model(self):
    self.P("Model setup initiated for {}".format(self.__class__.__name__))

    # Determine the device string based on configs and platform.
    self._setup_device_string()

    dev = th.device(self.get_device_string())
    if not self.has_device(dev):
      raise ValueError('Current machine does not have compute device {}'.format(dev))
    self.dev = dev

    if self.cfg_use_amp and self.cfg_fp16:
      self.P('Using both AMP and FP16 is not usually recommended.', color='r')

    # Now that we've chosen the device we can determine the default
    # backend order for this platform.
    self._platform_backend_priority = self._get_platform_backend_order()

    # record gpu info status pre-loading
    lst_gpu_status_pre = None
    if 'cuda' in self.get_device_string():
      lst_gpu_status_pre = self.log.gpu_info(mb=True)

    self.P("Prepping classes if available...")
    fn_classes = self.get_saved_classes_fn()
    url_classes = self.config_model.get(self.const.URL_CLASS_NAMES)
    if fn_classes is not None:
      self.download(
        url=url_classes,
        fn=fn_classes,
      )
      self.class_names = self.log.load_models_json(fn_classes)

    if self.class_names is None:
      self.P("WARNING: based class names loading failed. This is not necessarily an error as classes can be loaded within models.")

    if self.get_onnx_url is None and self.get_trt_url is None:
      # weights preparation - only for non-dynamic models
      # FIXME: this _seems_ to be only needed for torch (not ths) models and doesn't
      # match the rest of the code. Likely should be refactored to a _prepare_pytorch_model
      # and code should be pushed to users.
      if self.get_url is not None:
        self.P("Prepping model weights or graph...")
        model_weights = self.get_saved_model_fn()
        self.download(
          url=self.get_url,
          fn=model_weights,
        )
        # TODO: maybe use fn_weights as the output of self.download from above
        self.P("Loading model...")
        self.th_model = self._get_model(config=model_weights)
    else:
      backend_model_map={
        'ths' : (self.get_saved_model_fn(), self.get_url),
        'trt' : (self.get_model_trt_filename, self.get_trt_url),
        'onnx' : (self.get_model_onnx_filename, self.get_onnx_url),
        'openvino' : (self.get_model_onnx_filename, self.get_onnx_url),
      }
      self.th_model = self._get_model(config=backend_model_map)

    # Move the model to the required device if necessary.
    self._model_to(self.dev)

    # now load second stage model
    self._second_stage_model_load()

    # now warmup all models
    self._model_warmup()

    # record gpu info status post-loading
    taken, dev_nr = 0, 0
    if lst_gpu_status_pre is not None:
      dev_nr = 0 if dev.index is None else dev.index
      lst_gpu_status_post = self.log.gpu_info(mb=True)
      try:
        free_pre = lst_gpu_status_pre[dev_nr]['FREE_MEM']
        free_post = lst_gpu_status_post[dev_nr]['FREE_MEM']
        taken = free_pre - free_post
      except:
        self.P("Failed gpu check on {} PRE:{}, POST:{}\n {}\n>>>> END: GPU INFO ERROR".format(
          dev_nr, lst_gpu_status_pre, lst_gpu_status_post,
          self.trace_info()), color='error'
        )

    curr_model_dev = self._get_model_device()

    msg = "Model {} prepared on device '{}'{}:".format(
      self.server_name, curr_model_dev,
      " using {:.0f} MB GPU GPU {}".format(taken, dev_nr) if taken > 0 else "",
      )
    for k in self.config_model:
      msg = msg + '\n  {:<17} {}'.format(k+':', self.config_model[k])
    self.P(msg, color='g')
    return msg

  def _th_device_eq(self, device1, device2):
    """
    Compares two torch devices for equality.

    Parameters
    ----------

    device1: th.device, left hand side of the comparison
    device2: th.device, right hand side of the comparison

    Returns
    -------
    True iff the devices are equal, false otherwise
    """
    if ((device1.type != device2.type) or  # cpu vs cuda
        ((device1.type == device2.type) and
         (device1.index != device2.index)) # cuda but on different index
        ):
      return False
    return True

  def _get_model_device(self, model=None):
    """
    Returns the torch device assiciated with the model. If the model is
      None then the fist stage serving model is used.

    Parameters
    ----------

    model: torchscript, torch model or a ModelBackendWrapper.

    Returns
    -------

    ret - th.device - the corresponding torch device where the model is located.
    """
    if model is None:
      model = self.th_model
    if isinstance(model, ModelBackendWrapper):
      return model.get_device()
    return next(model.parameters()).device

  def _get_input_dtype(self, index, model=None):
    """
    Returns the torch dype for an input of the given model. If the model is
      None then the fist stage serving model is used.

    Parameters
    ----------

    index: int, the index of the paramter being queried.
    model: torchscript, torch model or a ModelBackendWrapper - The model
      for which we're querying the input parameter type. If the model is
      None then the fist stage serving model is used.

    Returns
    -------

    ret - th.dtype, the associated type of the input
    """
    if model is None:
      model = self.th_model
    if isinstance(model, ModelBackendWrapper):
      return model.get_input_dtype(index)
    return next(model.parameters()).dtype

  def _model_to(self, device, model=None):
    """
    Move a model to the specified torch device, converting it to fp16 if
    required (the serving configuration has USE_FP16 set to True). If the
    model is None then the fist stage serving model is used.

    Parameters
    ----------

    device: th.device, the device to move the model to

    model: torchscript, torch model or a ModelBackendWrapper - The model
      to move or conver. If the model is None then the fist stage serving
      model is used.

    Returns
    -------
    None
    """
    if model is None:
      model = self.th_model
    model_dev = self._get_model_device(model)
    if isinstance(model, ModelBackendWrapper):
      # The should already be on the correct device. If it's not
      # then something has gone horribly wrong.
      if not self._th_device_eq(model_dev, device):
        raise ValueError("Backend model of type {} on {} should already be on device: {}".format(
            model.__class__.__name__,
            model_dev,
            device
          )
        )
      return
    #endif backend model
    if not self._th_device_eq(model_dev, device):
      self.P("Model '{}' loaded & placed on '{}' -  moving model to device:'{}'".format(
        self.__class__.__name__, model_dev, device), color='y')
      model.to(device)
    model.eval()
    if self.cfg_fp16:
      model.half()
    return

  def _process_config_classes(self, config):
    config['names'] = {int(k): v for k, v in config.get('names', {}).items()}
    class_names = self._class_dict_to_list(config['names'])
    if len(class_names) > 0:
      self.P("Loading class names : {} classes ".format(len(class_names)))
      self.class_names = class_names
    return config

  def download_model_for_backend(self, url, fn_model, backend):
    """
    Downloads a model for a backend from the specifed url to:

       <models-folder>/backend/fn_model{no_suffix}/fn_model

    This will additionally create folders if needed.

    Parameters
    ----------
      url - str, The URL used for download
      fn_model - str, the downloaded model file
      backend - the name of the backend

    Returns
    -------
    ret - str - the path to the the downloaded model folder
    """
    if url is None:
      raise ValueError(f"No URL specified for model {fn_model} for backend {backend}")
    from pathlib import Path
    import os

    fn_model_no_suffix = str(Path(fn_model).with_suffix(''))
    relative_backend_folders = os.path.join(backend, fn_model_no_suffix)
    model_dir = os.path.join(self.log.get_models_folder(), relative_backend_folders)
    # Make sure folders are in fact there.
    os.makedirs(model_dir, exist_ok=True)
    fn_path = os.path.join(model_dir, fn_model)
    download_path = os.path.join(relative_backend_folders, fn_model)
    self.download(url, download_path)
    if not os.path.isfile(fn_path):
      raise ValueError(f"Could not download model file from {url}. {fn_model} not found.")
    return model_dir

  def prepare_model(
    self,
    backend_model_map : dict = None,
    forced_backend : str = None,
    post_process_classes : bool = False,
    return_config : bool = False,
    **kwargs
  ):
    """
    Creates the optimal model for this platform according
    to the backend_model_map parameter. The model will be downloaded
    (if required), saved to disk (if required) and loaded.
    Adds the model configuration (dict) to serving graph_config
    dictionary, with the key being the selected model filename.

    Raises a RuntimeError if no model was selected.

    Examples:
      model, config, fn_model = self.prepare_model(
        backend_model_map={
          'ths' : (self.get_saved_model_fn(), self.get_url),
          'trt' : (self.get_model_trt_filename, self.get_trt_url),
          'onnx' : (self.get_model_onnx_filename, self.get_onnx_url),
          'openvino' : (self.get_model_onnx_filename, self.get_onnx_url),
        },
        post_process_classes=True,
        return_config=True
      )
      # Creates a torchscript model.
      model = self.prepare_model(
        backend_model_map={
          'ths' : (self.get_saved_model_fn(), self.get_url),
          'trt' : (self.get_model_trt_filename, self.get_trt_url),
        },
        forced_backend='ths',
        post_process_classes=True,
        return_config=False
      )

    Parameters
    ----------
    backend_model_map : dict
      Keys are backend names and values are tuples of (filename, url)

    forced_backend : str | None
      If not None the method will use the provided backend instead of
      selecting one according to the platform configuration.

    post_process_classes : bool
      Updates the returned configuration, changing keys from strings
      to integers

    return_config : bool
      If True, returns a tuple of (model, model configuration, filename)
      If False, only returns the model.

    Returns
    -------
    If return_config is False:
      res - torchscript model | BackendModelWrapper
    If return_config is True:
      res - tuple of (torchscript model | BackendModelWrapper,
                      model configuration,
                      selected model filename)
    """

    backend_order = self._platform_backend_priority
    if forced_backend is not None:
      backend_order = [forced_backend]

    for backend in backend_order:
      if backend == 'trt':
        load_method = self._prepare_trt_model
      if backend == 'onnx':
        load_method = self._prepare_onnx_model
      if backend == 'openvino':
        load_method = self._prepare_openvino_model
      if backend == 'ths':
        load_method = self._prepare_ts_model

      if backend_model_map is not None:
        # If the backend model map was passed don't fall back to first stage
        # defaults. This is to support the case where we want to load a second
        # stage model, and for this case falling back to the first stage model
        # would be wrong.
        fns = backend_model_map.get(backend)
        if fns is None:
          self.P("No backend map entry found for backend {}, skipping".format(backend))
          continue
        fn_model, fn_url = fns
        if fn_model is None:
          self.P("No model found for backend {}, skipping".format(backend))
          continue
      #endif check for non-default backend model map

      self.P("Loading for backend {}".format(backend))
      model, config = load_method(
        url=fn_url,
        fn_model=fn_model,
        post_process_classes=post_process_classes,
        return_config=True,
        **kwargs
      )

      if model is not None:
        return (model, config, fn_model) if return_config else model
      #endif check if model loaded
    #endfor backends in order

    raise RuntimeError("Could not prepare model")
    return

  def get_input_shape(self):
    return (3, *self.cfg_input_size) if self.cfg_input_size is not None else None

  @staticmethod
  def __model_call_callback(th_inputs, model):
    """
    This is the default method for calling a model with the given inputs.
    Parameters
    ----------
    th_inputs - the inputs to be passed to the model
    model - the model to be called

    Returns
    -------
    res - the result of the model call
    """
    return model(th_inputs)

  def get_model_call_method(self, model_call_method=None):
    """
    This method returns the model call method to be used for calling the model.
    Parameters
    ----------
    model_call_method - the model call method to be used for calling the model

    Returns
    -------
    model_call_method - the model call method to be used for calling the model
    """
    if model_call_method is None:
      # If the method is not provided a custom method will be checked
      # and if it is not found the default method will be used
      default_method = getattr(self, 'model_call', None)
      return default_method if callable(default_method) else self.__model_call_callback
    return model_call_method

  def _forward_pass(self, th_inputs, model=None, model_call_method=None, debug=None, debug_str='', autocast=True):
    model = self.th_model if model is None else model
    debug = self._full_debug if debug is None else debug
    model_call_method = self.get_model_call_method(model_call_method=model_call_method)
    # endif model not provided
    if autocast:
      with th.cuda.amp.autocast(enabled=self.cfg_use_amp):
        with th.no_grad():
          if debug:
            self.P("  Forward pass {}, dev '{}' with {}:{}".format(
              debug_str, th_inputs.device, th_inputs.shape, th_inputs.dtype
            ))
          th_preds = model_call_method(model=model, th_inputs=th_inputs)
        # end no_grad
      # end autocast
    else:
      with th.no_grad():
        if debug:
          self.P("  Forward pass {}, no autocast on dev '{}' with {}:{}".format(
            debug_str, th_inputs.device, th_inputs.shape, th_inputs.dtype
          ))
        th_preds = model_call_method(model=model, th_inputs=th_inputs)
      # end no_grad
    # endif autocast or not
    return th_preds

  # TODO: maybe add support for randint?
  def model_warmup_helper(
      self, model=None, input_shape=None, warmup_count=None,
      max_batch_size=None, model_dtype=None, model_device=None,
      model_name=None, model_call_method=None
  ):
    """
    This method is used to warm up the model.
    Parameters
    ----------
    model - the model to be warmed up
    input_shape - the shape of the input to be used for the model after warmup
    warmup_count - the number of forward passes per batch size to be used for warmup
    max_batch_size - the maximum batch size to be used for the model after warmup
    model_dtype - the dtype of the input to be used for the model after warmup
    model_device - the device of the input to be used for the model after warmup
    model_name - the name of the model

    Returns
    -------
    None
    """
    model = self.th_model if model is None else model
    input_shape = self.get_input_shape() if input_shape is None else input_shape
    warmup_count = self.cfg_warmup_count if warmup_count is None else warmup_count
    max_batch_size = self.cfg_max_batch_first_stage if max_batch_size is None else max_batch_size
    model_name = self.__class__.__name__ if model_name is None else model_name

    if model is not None:
      if model_dtype is None or model_device is None:
        model_dtype = self._get_input_dtype(0, model=model)
        model_device = self._get_model_device(model=model)
      # endif model_dtype or model_device
      model_dev_type = model_device.type
      if input_shape is not None:
        self.P(f"Warming up model {model_name} with {input_shape}/{model_dtype} on device '{model_device}'"
               f" for MAX_BATCH_SIZE {max_batch_size} and WARMUP_COUNT {warmup_count}...")
        for wbs in range(1, max_batch_size + 1):
          shape = (wbs, *input_shape)
          th_warm = th.rand(
            *shape,
            device=model_device,
            dtype=model_dtype
          )
          for warmup_pass in range(1, warmup_count + 1):
            _ = self._forward_pass(
              th_warm, model=model, model_call_method=model_call_method,
              autocast='cuda' in model_dev_type and self.cfg_use_amp,
              debug=self._full_debug, debug_str=str(warmup_pass)
            )
          # endfor warmup_pass
        # endfor warmup_batch_size
      self.P("Model {} warmed up and ready for inference".format(model_name))
    else:
      self.P("ERROR: Model of {} not found".format(model_name))
    return

  def _model_warmup(self):
    shape = self.get_input_shape()
    th.backends.cudnn.benchmark = self.cfg_cudnn_benchmark

    self.model_warmup_helper(
      model=self.th_model,
      input_shape=shape,
      warmup_count=self.cfg_warmup_count,
      max_batch_size=self.cfg_max_batch_first_stage
    )

    self._second_stage_model_warmup()
    return

  def _class_dict_to_list(self, dct_classes):
    keys = sorted(list(dct_classes.keys()))
    result = [dct_classes[x] for x in keys]
    return result

  def ver_element_to_int(self, ver_element):
    """
    Method for getting rid of non-numeric suffixes from version elements.
    For example if our current version is '0.15.2+cpu' and the minimum version is '0.15.2' we want to
    still be able to tell that the current version is greater than or equal to the minimum version.
    Parameters
    ----------
    ver_element - the version element to be converted to int

    Returns
    -------
    int - the numeric value of the version element or 0 if the version element does not start with a digit.
    """
    ver_element_numeric = self.re.sub('\D.*', '', ver_element)
    return int(ver_element_numeric) if ver_element_numeric.isnumeric() else 0

  def ver_to_int(self, version):
    ver_elements = version.split('.')
    int_ver_elements = [self.ver_element_to_int(x) for x in ver_elements[:3]]
    weights = [1000000, 1000, 1]
    int_ver = sum(int_ver_elements[i] * weights[i] for i in range(min(len(weights), len(int_ver_elements))))

    return int_ver

  def check_version(self, min_ver, curr_ver):
    return self.ver_to_int(min_ver) <= self.ver_to_int(curr_ver)

  def valid_version(self, ver_str):
    return '.' in ver_str

  def check_versions(self, model_config, fn_path, env_versions, err_keys):
    versions = {key: model_config[key] if key in model_config.keys() else 'Unspecified' for key in env_versions.keys()}

    err_check = not all([
      self.check_version(min_ver=versions[vkey], curr_ver=env_versions[vkey]) and self.valid_version(versions[vkey])
      for vkey in err_keys
    ])
    warn_check = not all([
      self.valid_version(versions[vkey]) and self.check_version(min_ver=versions[vkey], curr_ver=env_versions[vkey])
      for vkey in env_versions.keys()
    ])

    if err_check:
      err_msg = f'ERROR! Model from {fn_path} has versions above current environment versions!' \
                f'[Model versions:{versions} > Env versions: {env_versions}]'
      self.P(err_msg, color='e')
      raise Exception(err_msg)
    elif warn_check:
      warn_msg = f'WARNING! Model from {fn_path} has versions above current environment versions or has ' \
                 f'unspecified/invalid versions! [Model versions:{versions} > Env versions: {env_versions}]'
      self.P(warn_msg, color='r')
    else:
      self.P(f'Graph version check passed! [{versions} <= {env_versions}]')
    # endif version_check
    return

  def _get_model(self, config):
    """
    Get the model associated with configuration config.

    Parameters
    ----------
    config : str | dict
      If config is a dict, keys are backend names and values are tuples
        of (filename, url).
      Current valid backend names are:
        'ths'      - torchscript backend
        'onnx'     - ONNX backend
        'openvino' - OpenVINO backend
        'trt'      - TensorRT backend
      If config is a string, this is the torchscript file name.
        Note this is used for backwards compatibility and is deprecated.
        Please use a dict parameter.

    Returns
    -------
    res - torch/torchscript model | ModelBackendWrapper
    """
    raise NotImplementedError()

  def _pre_process_images(self, images):
    raise NotImplementedError()

  # second stage methods
  def _second_stage_model_load(self):
    # this method defines default behavior for second stage model load
    return

  def _second_stage_model_warmup(self):
    # this method defines default behavior for second stage model warmup
    return

  def _second_stage_process(self, th_preds, th_inputs=None, **kwargs):
    # this function processed post forward and should be used in subclasses
    # also here you we can add second/third stage classifiers/regressors (eg mask detector)
    return th_preds

  def _second_stage_classifier(self, first_stage_out, th_inputs):
    # override this to add second stage classifier/regressor fw pass
    # for this to be executed in model prep set `self.has_second_stage_classifier = True`
    return None

  # end second stage methods

  @staticmethod
  def _th_resize(self, lst_images, target_size):
    return

  ###
  ### BELOW MANDATORY (or just overwritten) FUNCTIONS:
  ###

  def _get_inputs_batch_size(self, inputs):
    """
    Extract the batch size of the inputs. This method is can be overwritten when inputs are dictionaries,
    because `len(inputs)` would simply be the number of keys in the dict.

    Example of use case: in `th_cqc`, the input is a kwarg dict with images and anchors, so the batch size
    should be the len of those inputs, not the number of inputs (which is constantly 2).

    Parameters
    ----------
    inputs : Any
        The input of the serving model

    Returns
    -------
    int
        The batch size of the input
    """
    return len(inputs) # not the greatest method but it works

  def _startup(self):
    msg = self._setup_model()
    return msg

  def _pre_process(self, inputs):
    """
    This method does only the basic data extraction from upstream payload and
    has replaced:
      ```
      def _pre_process(self, inputs):
        prep_inputs = th.tensor(inputs, device=self.dev, dtype=self.th_dtype)
        return prep_inputs
      ```


    Parameters
    ----------
    inputs : list
      list of batched numpy images.

    Raises
    ------
    ValueError
      DESCRIPTION.

    Returns
    -------
    TYPE
      DESCRIPTION.

    """
    lst_images = None
    if isinstance(inputs, dict):
      lst_images = inputs.get('DATA', None)
    elif isinstance(inputs, list):
      lst_images = inputs

    if lst_images is None or len(lst_images) == 0:
      msg = "Unknown or None `inputs` received: {}".format(inputs)
      self.P(msg, color='error')
      raise ValueError(msg)
    return self._pre_process_images(lst_images)

  def _aggregate_batch_predict(self, lst_results):
    if len(lst_results) == 1:
      return lst_results[0]
    else:
      if isinstance(lst_results[0], th.Tensor):
        return th.cat(lst_results, dim=0)
      else:
        raise NotImplementedError(
          "Please implement `_aggregate_batch_predict` method that can aggregate model output type: {}".format(
            type(lst_results[0]))
        )
    return

  def _batch_predict(self, prep_inputs, model, batch_size, aggregate_batch_predict_callback=None, **kwargs):
    lst_results = []
    start_time = self.time()

    if isinstance(prep_inputs, dict):
      last_dim = None
      for key in prep_inputs.keys():
        assert isinstance(prep_inputs[key], (list, th.Tensor, self.np.ndarray)), 'Invalid input type {}'.format(type(prep_inputs[key]))
        assert last_dim is None or last_dim == len(prep_inputs[key]), 'Inconsistent input dim: {} vs {}'.format(last_dim, prep_inputs[key])
        last_dim = len(prep_inputs[key])
      #endfor
      real_batch_size = last_dim
    else:
      real_batch_size = len(prep_inputs)
    #endif

    if batch_size is None:
      self.maybe_log_phase('_model_predict[bs=None]', start_time, done=False)
      if isinstance(prep_inputs, dict):
        return model(**prep_inputs, **kwargs)
      else:
        return model(prep_inputs, **kwargs)
      #endif
    #endif

    self.maybe_log_phase(f'_model_predict[bs={real_batch_size} mbs={batch_size} and bn={self.np.ceil(real_batch_size / batch_size).astype(int)}]', start_time, done=False)
    num_batches = self.np.ceil(real_batch_size / batch_size).astype(int)
    for i in range(num_batches):
      if isinstance(prep_inputs, dict):
        batch = {}
        for key in prep_inputs.keys():
          batch[key] = prep_inputs[key][i * batch_size: (i+1) * batch_size]
        # endfor
        th_inferences = model(**batch, **kwargs)
      else:
        batch = prep_inputs[i * batch_size:(i+1) * batch_size]
        th_inferences = model(batch, **kwargs)
      # endif

      lst_results.append(th_inferences)
    # endfor batches
    self.maybe_log_phase(f'_model_predict[bs={real_batch_size} mbs={batch_size} and bn={self.np.ceil(real_batch_size / batch_size).astype(int)}]', start_time)
    if aggregate_batch_predict_callback is None:
      return self._aggregate_batch_predict(lst_results)
    else:
      return aggregate_batch_predict_callback(lst_results)

  def _predict(self, prep_inputs):
    ### TODO: This no longer works with multiple inputs
    if isinstance(prep_inputs, dict):
      kwargs_first_stage = prep_inputs.get('kwargs_first_stage', {})
      kwargs_second_stage = prep_inputs.get('kwargs_second_stage', {})
      inputs = prep_inputs.get('inputs', None)
      if inputs is None:
        raise ValueError('Prepocess returned dict without inputs')
    elif not isinstance(prep_inputs, (list, th.Tensor, self.np.ndarray)):
      raise ValueError('Preprocess returned invalid type {}'.format(type(prep_inputs)))
    else:
      inputs = prep_inputs
      kwargs_first_stage = {}
      kwargs_second_stage = {}
    #endif

    start_time = self.time()
    bs = self._get_inputs_batch_size(inputs)
    with th.cuda.amp.autocast(enabled=self.cfg_use_amp):
      with th.no_grad():
        str_predict_timer = 'fwd_b{}_{}'.format(str(bs), str(self.cfg_max_batch_first_stage))
        self.log.start_timer(self.predict_server_name + '_' + str_predict_timer)
        self.maybe_log_phase('_batch_process', start_time, done=False)
        th_x_stage1 = self._batch_predict(
          prep_inputs=inputs,
          model=self.th_model,
          batch_size=self.cfg_max_batch_first_stage,
          **kwargs_first_stage
        )
        self.maybe_log_phase('_batch_predict', start_time)
        self.log.stop_timer(self.predict_server_name + '_' + str_predict_timer)

        self._start_timer('stage2')
        self.maybe_log_phase('_second_stage_process', start_time, done=False)
        th_x_stage2 = self._second_stage_process(
          th_preds=th_x_stage1,
          th_inputs=inputs,
          **kwargs_second_stage
        )
        self.maybe_log_phase('_second_stage_process', start_time)

        self._stop_timer('stage2')
    return th_x_stage2

  def _post_process(self, preds):
    if isinstance(preds, (list, tuple)):
      # lets unpack
      outputs = []
      for pred in preds:
        if isinstance(pred, th.Tensor):
          outputs.append(pred.cpu().numpy())
    elif isinstance(preds, th.Tensor):
      outputs = preds.cpu().numpy()
    else:
      raise ValueError("Unknown model preds output of type {}".format(type(preds)))
    return outputs

  def _clear_model(self):
    if hasattr(self, 'th_model') and self.th_model is not None:
      self.P("Deleting current loaded model for {} ...".format(self.__class__.__name__))
      curr_model_dev = self._get_model_device(self.th_model)
      if 'cuda' in curr_model_dev.type.lower():
        lst_gpu_status_pre = self.log.gpu_info(mb=True)

      del self.th_model
      self.th_utils.clear_cache()

      freed = 0
      idx = None
      if 'cuda' in curr_model_dev.type.lower():
        lst_gpu_status_post = self.log.gpu_info(mb=True)
        try:
          idx = 0 if curr_model_dev.index is None else curr_model_dev.index
          freed = lst_gpu_status_post[idx]['FREE_MEM'] - lst_gpu_status_pre[idx]['FREE_MEM']
        except:
          self.P("Failed gpu check on {} PRE:{}, POST:{}, {}".format(
            idx, lst_gpu_status_pre[idx], lst_gpu_status_post[idx],
            self.trace_info()), color='error'
          )

      self.P("Cleanup finished. Freed {:.0f} MB on GPU {}".format(freed, idx))
    return

  def _shutdown(self):
    self._clear_model()
    return

  def forward(self, inputs):
    prep_inputs = self._pre_process(inputs)
    fwd = self.th_model(prep_inputs)
    return fwd
